//
//  BFRecordScreenController.swift
//  BFRecordScreenKit_Example
//
//  Created by 胡志强 on 2021/11/23.
//  Copyright © 2021 CocoaPods. All rights reserved.
//

import AVFAudio
import AVFoundation
import BFCommonKit
import BFMaterialKit
import BFMediaKit
import BFNetRequestKit
import BFUIKit
import Foundation
import GPUImage
import Photos
import UIKit

struct WithDrawModel {
    var type: Int // 0:拖动； 1:预览播放暂停 2: 录音结束  3: 删除录音
    var timestamp: CMTime
    var deletedVoices: [(PQVoiceModel, Int)]?
    // add by ak 保存删除的字幕数据用于恢复
    var deletedTittles: [PQEditSubTitleModel]?
}

public class BFRecordScreenController: BFBaseViewController {
    public var nextActionHandle: (() -> Void)?
    public var closeActionHandle: (() -> Void)?
    public var changeItemHandle: ((_ index: Int) -> Void)?
    // 当前录制结束的上报
    public var recordEndHandle: ((_ currentRecord: PQVoiceModel?) -> Void)?
    // 字幕按钮点击上报
    public var subTitleBtnClickHandle: ((_ isOn: Bool) -> Void)?
//    public var currentRecordId: String? // 当前录制Id

    // MARK: - 素材参数
    lazy var rsvmanager : BFRecordScreenVideoManager = {
        let m = BFRecordScreenVideoManager()
        m.assetPlayer = assetPlayer
        m.recordPlayer = recordPlayer
        return m
    }()
    
    lazy var rsimanager : BFRecordScreenImageManager = {
        let m = BFRecordScreenImageManager()
        m.recordPlayer = recordPlayer
        return m
    }()
    
    var rscmanager = BFRecordScreenCameraManager()
    var rscurrentManager = BFRecordScreenBaseManager()
    
    // MARK: - 录制参数
    public var assets = [PHAsset]()
    var currItemModelIndex = -1 {
        didSet{
            if currItemModelIndex < 0 {
                return
            }
            let itemModel = itemModels[currItemModelIndex]
            
            switch itemModel.mediaType {
            case .Image:
                rscurrentManager = rsimanager
                recordBtn.setTitle("按住 录制", for: .normal)
                progressThumV.isHidden = true
                
            case .Video:
                rscurrentManager = rsvmanager
                recordBtn.setTitle("按住 录音", for: .normal)
                progressThumV.isHidden = false

            case .Camera:
                rscurrentManager = rscmanager
                recordBtn.setTitle("按住 录音", for: .normal)
                progressThumV.isHidden = false
            
            default:
                break
            }
            rscurrentManager.dele = self
            if let cell = collectionView.cellForItem(at: IndexPath(item: currItemModelIndex, section: 0)) as? BFImageCoverViewCell {
                if (cell as? BFCameraCoverViewCell) != nil {
                    cell.playView.setInputRotation(GPUImageRotationMode(rawValue: 2), at: 0)
                    progressThumV.isHidden = true
                    if let rendV = (rscurrentManager as? BFRecordScreenCameraManager)?.rendView {
                        rendV.removeFromSuperview()
                        cell.playView.addSubview(rendV)
                        rendV.snp.makeConstraints { make in
                            make.edges.equalToSuperview()
                        }
                        
                        rendV.recordEndCallBack = {[weak self] isSuccess, sticker in
                            if let sticker = sticker {
                                itemModel.videoStickers.append(sticker)
                            }
                        }
                    }
                    
                }else {
                    cell.playView.setInputRotation(GPUImageRotationMode(rawValue: 0), at: 0)
                    progressThumV.isHidden = false
                    if itemModel.mediaType == .Video {
                        rsvmanager.assetPlayer
                    }
                }
                rscurrentManager.playView = cell.playView
                rscurrentManager.recordItem = itemModel
                rscurrentManager.resetEnv()
            }

        }
    }
    public var itemModels = [BFRecordItemModel]()
    // add by ak 当前的显示的字幕位置
    var showSubtitleIndex = -1
    //    var shouldPlayRecordIndex:Int = -1          // 当前应该播放的录音资源序号
    var currentPlayRecordIndex: Int = -1 // >= 0 :当前正在播放的录音资源序号； -3： 刚录音完，不需要播放录音； -1:初始化阶段
    public var displayLink: CADisplayLink? // 图片素材播放控制
    var isRecording = false { // 是否正在录音
        didSet {
            withDrawBtn.isHidden = isRecording
            changeVoiceBtn.isHidden = isRecording
            subtitleBtn.isHidden = isRecording
            soundSettingBtn.isHidden = isRecording

            recordBtn.setTitle(isRecording ? "松手 暂停" : (rscurrentManager.recordItem?.mediaType == .Camera ? "按住 录制" : "按住 录音"), for: .normal)
            recordBtn.backgroundColor = UIColor.hexColor(hexadecimal: "#389AFF", alpha: isRecording ? 0.6 : 1)
//            if !isRecording {
//                BFLog(1, message: "stop")
//            }
        }
    }

    var isEndPlay = false

    var isNormalPlaying = false { // 是否正在播放
        didSet {
            withDrawBtn.isHidden = isNormalPlaying
        }
    }

    var currentAssetProgress: CMTime = .zero
    {
        didSet {
            BFLog(3, message: "currentAssetProgress=\(currentAssetProgress.seconds)")
            rscurrentManager.currentAssetProgress = currentAssetProgress
        }
    } // 当前素材播放的进度
    // 播放器开始播放时间
    var recordStartPlayTime: CMTime = .zero
    // 某个录音开始播放时间
    var currenStartPlayTime: CMTime = .zero
    var pauseTime: Double = 0 // 停止无操作的时间点

    var assetPlayer: AVPlayer = AVPlayer(playerItem: nil) // 原视频音频播放器
    var recordPlayer: AVPlayer = AVPlayer(playerItem: nil) // 录音音频播放器

    var hadPrepareToPlayRecord = false // 录音播放器准备
    var playView: GPUImageView? // 视频展示视图

    // MARK: 行为参数

    var movieIsProcessing = false

    // 行为记录，方便撤销
    var events = [WithDrawModel]() {
        didSet {
            withDrawBtn.isEnabled = (events.count != 0)
        }
    }

    // 是否在拖动进度条
    var isDragingProgressSlder: Bool = false

    var isStopAtRecordRange = -1

    // 保存识别出来的字幕信息,用于回放，和合成使用
    //    var saveSubTitles:[PQEditSubTitleModel] = Array.init()

    // MARK: - 录音相关

    // 定义音频的编码参数
    let recordSettings: [String: Any] = [AVSampleRateKey: 44100.0, // 声音采样率
                                         AVFormatIDKey: kAudioFormatLinearPCM, // 编码格式
                                         AVNumberOfChannelsKey: 1, // 采集音轨
                                         AVEncoderBitDepthHintKey: 16, // 位深
                                         AVEncoderAudioQualityKey: AVAudioQuality.medium.rawValue] // 音频质量

    public var haveSpeakVolume: Float = 0.15
    public var noSpeakVolume: Float = 0.30

    // 录音管理器
    var recorderManager: BFVoiceRecordManager?

    var NeoNuiAPPID: String?
    var NeoNuiToken: String?

    // MARK: - 视图参数

    var beginOnStartBtn: Bool = false
    var touchStart: CGPoint = CGPoint(x: 0, y: 0)
    var avplayerTimeObserver: NSKeyValueObservation?
    var recordPlayerTimeObserver: NSKeyValueObservation?

    var indirectionView: BFIndirectionProgressView?

    lazy var progreddL: UILabel = {
        let l = UILabel()
        l.textAlignment = .center
        l.font = UIFont.systemFont(ofSize: 13, weight: .medium)
        l.textColor = .white
        l.layer.shadowOpacity = 0.4
        l.layer.shadowColor = UIColor.black.cgColor
        l.layer.shadowOffset = CGSize(width: 0, height: 2)
        l.layer.shadowRadius = 2
        l.text = "00:00"
        return l
    }()

    lazy var playBtn: UIButton = {
        let btn = UIButton(frame: CGRect(x: 0, y: 0, width: 150, height: 156))
        btn.setImage(imageInRecordScreenKit(by: "preview_play"), for: .normal)
        let vv = UIView(frame: CGRect(x: 0, y: 0, width: 0.5, height: 0.5))
        vv.backgroundColor = UIColor.init(white: 1, alpha: 0.3)
        btn.setImage(vv.graphicsGetImage(), for: .selected)
        btn.addTarget(self, action: #selector(playVideo(btn:)), for: .touchUpInside)
        btn.addTarget(self, action: #selector(playbtnDown(btn:)), for: .touchDown)
        btn.addTarget(self, action: #selector(playbtnCancle(btn:)), for: .touchCancel)
        btn.center = view.center
        return btn
    }()

    lazy var bottomeView: UIImageView = {
        let iv = UIImageView(image: imageInRecordScreenKit(by: "bottom_shadow"))
        iv.contentMode = .scaleAspectFill
        iv.isUserInteractionEnabled = true
        // 拖曳手勢
        let pan = UIPanGestureRecognizer(target: self, action: #selector(pan(recognizer:)))
        pan.delegate = self
        pan.minimumNumberOfTouches = 1
        pan.maximumNumberOfTouches = 1
        iv.addGestureRecognizer(pan)

        return iv
    }()

    lazy var recordBtn: UIButton = {
        let btn = UIButton(type: .custom)
        btn.backgroundColor = ThemeStyleColor
        btn.setTitle("按住 录音", for: .normal)
        btn.addCorner(corner: 6)
        btn.titleLabel?.font = UIFont.boldSystemFont(ofSize: 18)
        btn.adjustsImageWhenHighlighted = false
        btn.addTarget(self, action: #selector(startRecord), for: .touchDown)
        btn.addTarget(self, action: #selector(endRecord), for: .touchUpInside)
        btn.isEnabled = false
        return btn
    }()

    lazy var deleteRecordBtn: UIButton = {
        let btn = UIButton(type: .custom)

        btn.backgroundColor = UIColor.hexColor(hexadecimal: "#FF0000", alpha: 0.3)
        btn.setTitle("删除录制", for: .normal)
        btn.adjustsImageWhenHighlighted = false
        btn.titleLabel?.font = UIFont.boldSystemFont(ofSize: 18)
        btn.addCorner(corner: 6)
        btn.addTarget(self, action: #selector(deleteRecordAction), for: .touchUpInside)
        btn.isHidden = true
        return btn
    }()

    lazy var withDrawBtn: UIButton = {
        let btn = UIButton(type: .custom)
        btn.setImage(imageInRecordScreenKit(by: "withdraw_n"), for: .normal)
        btn.setImage(imageInRecordScreenKit(by: "withdraw_h"), for: .highlighted)
        btn.setTitle("回退", for: .normal)
        btn.setTitleColor(.white, for: .normal)
        btn.setTitleColor(.lightGray, for: .disabled)
        btn.isEnabled = false
        btn.titleLabel?.font = UIFont.systemFont(ofSize: 12)
        btn.contentVerticalAlignment = UIControl.ContentVerticalAlignment.center
        btn.addTarget(self, action: #selector(withdrawAction), for: .touchUpInside)
        return btn
    }()

    lazy var changeVoiceBtn: UIButton = {
        let btn = UIButton(type: .custom)
        btn.setImage(imageInRecordScreenKit(by: "changeVoice_n"), for: .normal)
        btn.setImage(imageInRecordScreenKit(by: "changeVoice_h"), for: .highlighted)
        btn.setTitle("变声", for: .normal)
        btn.setTitleColor(.white, for: .normal)
        btn.setTitleColor(ThemeStyleColor, for: .highlighted)
        btn.titleLabel?.font = UIFont.systemFont(ofSize: 12)
        btn.contentVerticalAlignment = UIControl.ContentVerticalAlignment.center
        btn.addTarget(self, action: #selector(changeVoiceAction), for: .touchUpInside)
        return btn
    }()

    lazy var toolV: BFIntroduceToolView = {
        let toolV = BFIntroduceToolView()
        toolV.centerY = view.centerY

        toolV.choosedToolHandle = { [weak self] tool in
            guard let sself = self else {
                return
            }
            tool.center = sself.view.center
            sself.view.addSubview(tool)
        }

        return toolV

    }()

    // 头像  add by ak
    lazy var avatarView: BFRecordAvatarView = {
        let avatarView = BFRecordAvatarView(frame: CGRect(x: 10, y: 10, width: 120, height: 120))
        avatarView.isHidden = true
        return avatarView
    }()

    // 打开摄像头
    lazy var openCameraBtn: UIButton = {
        let btn = UIButton(type: .custom)
        btn.setImage(imageInRecordScreenKit(by: "openCamera"), for: .normal)
        btn.addTarget(self, action: #selector(openCamera), for: .touchUpInside)
        return btn
    }()

    // 画笔
    lazy var drawPinBtn: UIButton = {
        let btn = UIButton(type: .custom)
        btn.setImage(imageInRecordScreenKit(by: "drawPin"), for: .normal)
        btn.addTarget(self, action: #selector(drawPin), for: .touchUpInside)
        return btn
    }()

    // 字幕设置
    lazy var subtitleBtn: UIButton = {
        let btn = UIButton(type: .custom)
        btn.setImage(imageInRecordScreenKit(by: "subtitleBtn_on"), for: .normal)
        btn.addTarget(self, action: #selector(subTitleClick), for: .touchUpInside)
        return btn
    }()

    // 声音设置
    lazy var soundSettingBtn: UIButton = {
        let btn = UIButton(type: .custom)
        btn.setImage(imageInRecordScreenKit(by: "soundBtn"), for: .normal)
        btn.addTarget(self, action: #selector(soundSetting), for: .touchUpInside)
        return btn
    }()

    // 字幕设置面板
    lazy var subtitleSettingView: BFSubtitleSettingView = {
        let subtitleSetting = BFSubtitleSettingView(frame: CGRect(x: 0, y: 0, width: cScreenWidth, height: cScreenHeigth))
        subtitleSetting.isHidden = true
        return subtitleSetting
    }()

    // 编辑字幕
    lazy var subtitleEditView: BFSubtitleEditView = {
        let subtitleEditView = BFSubtitleEditView(frame: CGRect(x: 0, y: 0, width: cScreenWidth, height: cScreenHeigth))
        subtitleEditView.isHidden = true

        return subtitleEditView
    }()

    // 显示字幕
    lazy var subtitleLabel: PQSubTitleLabel = {
        let subtitleLabel = PQSubTitleLabel()
        subtitleLabel.tag = 100_100
        subtitleLabel.textAlignment = .center
        subtitleLabel.numberOfLines = 0
        subtitleLabel.lineBreakMode = .byWordWrapping
        subtitleLabel.isUserInteractionEnabled = true
        subtitleLabel.addGestureRecognizer(UITapGestureRecognizer(target: self, action: #selector(editSubtitle)))
        return subtitleLabel
    }()

    // 音量设置
    lazy var audioSettingView: BFAudioSettingView = {
        let audioSettingView = BFAudioSettingView(frame: CGRect(x: 0, y: 0, width: cScreenWidth, height: cScreenHeigth))
        audioSettingView.isHidden = true
        audioSettingView.haveSpeakSlider.valueIsInt = true
        audioSettingView.noSpeakSlider.valueIsInt = true
        audioSettingView.callBack = { [weak self] haveSpeak, noHaveSpeak in
            BFLog(1, message: "haveSpeak is:\(haveSpeak),noHaveSpeak is:\(noHaveSpeak)")
            self?.haveSpeakVolume = haveSpeak / 100.0
            self?.noSpeakVolume = noHaveSpeak / 100.0
            if !(self?.isNormalPlaying ?? false), !(self?.isRecording ?? false) {
                if self?.deleteRecordBtn.isHidden ?? false {
//                    self?.recordPlayer.volume = 0
                    self?.assetPlayer.volume = self?.noSpeakVolume ?? 1.0
                } else {
//                    self?.recordPlayer.volume = 1.0
                    self?.assetPlayer.volume = self?.haveSpeakVolume ?? 0.0
                }
            }
        }

        return audioSettingView
    }()

    lazy var progressThumV: BFVideoThumbProgressView = {
        let vv = BFVideoThumbProgressView(frame: CGRect(x: 0, y: 18, width: cScreenWidth, height: 50))
        vv.dragStartHandle = { [weak self] in
            self?.isDragingProgressSlder = true
            self?.pause()
        }
        vv.dragScrollProgressHandle = { [weak self] isStart, process in
            DispatchQueue.main.async { [weak self] in
                guard let sself = self else {
                    return
                }
                if isStart {
                    sself.events.append(WithDrawModel(type: 0, timestamp: sself.currentAssetProgress))
                }

                sself.isDragingProgressSlder = true

                sself.changeProgress(progress: process)
            }
        }
        vv.dragEndHandle = { [weak self] process in
            DispatchQueue.main.async { [weak self] in
                guard let sself = self else {
                    return
                }
                sself.thumbViewEnded(progress: process)
            }
        }
        vv.isHidden = true
        return vv
    }()

    lazy var collectionView: UICollectionView = {
        let flowLayout = UICollectionViewFlowLayout()
        flowLayout.minimumLineSpacing = 0
        flowLayout.minimumInteritemSpacing = 0
        flowLayout.scrollDirection = .horizontal
        flowLayout.itemSize = CGSize(width: view.frame.width, height: view.frame.height)
        let collectionView = UICollectionView(frame: CGRect(x: 0, y: 0, width: view.frame.width, height: view.frame.height), collectionViewLayout: flowLayout)
        collectionView.register(BFImageCoverViewCell.self, forCellWithReuseIdentifier: String(describing: BFImageCoverViewCell.self))
        collectionView.register(BFVideoCoverViewCell.self, forCellWithReuseIdentifier: String(describing: BFVideoCoverViewCell.self))
        collectionView.register(BFCameraCoverViewCell.self, forCellWithReuseIdentifier: String(describing: BFCameraCoverViewCell.self))
        collectionView.isPagingEnabled = true
        collectionView.bounces = false
        collectionView.showsVerticalScrollIndicator = false
        collectionView.showsHorizontalScrollIndicator = false
        collectionView.delegate = self
        collectionView.dataSource = self
        collectionView.backgroundColor = UIColor.clear
        if #available(iOS 11.0, *) {
            collectionView.contentInsetAdjustmentBehavior = .never
        } else {
            automaticallyAdjustsScrollViewInsets = false
        }
        return collectionView
    }()

    // debug
    lazy var neoNuiDebugLabel: UILabel = {
        let neoNuiDebugLabel = UILabel(frame: CGRect(x: 0, y: 100, width: 300, height: 24))
        neoNuiDebugLabel.backgroundColor = .black
        neoNuiDebugLabel.textColor = .white
        neoNuiDebugLabel.textAlignment = .left
        return neoNuiDebugLabel

    }()

    lazy var audioQueueRecoderLabel: UILabel = {
        let audioQueueRecoderLabel = UILabel(frame: CGRect(x: 0, y: 125, width: 300, height: 24))
        audioQueueRecoderLabel.backgroundColor = .black
        audioQueueRecoderLabel.textColor = .white
        audioQueueRecoderLabel.textAlignment = .left
        return audioQueueRecoderLabel

    }()

    // MARK: - ----------------- 生命周期

    deinit {
        rscmanager.clean()
        rsvmanager.clean()
        rsimanager.clean()
        
        NotificationCenter.default.removeObserver(self)
        avplayerTimeObserver?.invalidate()
        recordPlayerTimeObserver?.invalidate()
        if isRecording {
            recorderManager?.stopRecord(isCancel: true)
        }
        assetPlayer.pause()
        recordPlayer.pause()
    }

    @objc func didBecomeActive() {
        BFLog(message: "进入活跃状态")
    }

    @objc func willResignActive() {
        BFLog(message: "进入非活跃状态")
        endRecord()
    }

    /// 线路切换
    /// - Parameter nofify: <#nofify description#>
    @objc func routeChangeNofify(nofify: Notification) {
        let routeChangeDic = nofify.userInfo
        let routeChangeReason: AVAudioSession.RouteChangeReason? = AVAudioSession.RouteChangeReason(rawValue: UInt((routeChangeDic?[AVAudioSessionRouteChangeReasonKey] as? Int) ?? 0))
        if routeChangeReason == .oldDeviceUnavailable {
            let previousRoute = routeChangeDic?[AVAudioSessionRouteChangePreviousRouteKey] as? AVAudioSessionRouteDescription
            if previousRoute?.outputs.first?.portType == .headphones {
                playVideo(btn: playBtn)
            }
        }
    }

    override public func viewWillAppear(_ animated: Bool) {
        super.viewWillAppear(animated)
        navigationController?.isNavigationBarHidden = true
        hiddenNavigation()
    }
    
    public override func viewDidAppear(_ animated: Bool) {
        super.viewDidAppear(animated)
        endScrollItem(collectionView)
        // 设置默认值
        setSubtitleStyle(settingModel: subtitleSettingView.subtitle.setting)
    }

    override public func viewWillDisappear(_ animated: Bool) {
        super.viewWillDisappear(animated)
        pause()
    }

    override public func viewDidLoad() {
        super.viewDidLoad()
        _ = disablePopGesture()

        // 进入活跃状态
        PQNotification.addObserver(self, selector: #selector(didBecomeActive), name: UIApplication.didBecomeActiveNotification, object: nil)
        PQNotification.addObserver(self, selector: #selector(routeChangeNofify(nofify:)), name: AVAudioSession.routeChangeNotification, object: nil)
        // 进入非活跃状态
        PQNotification.addObserver(self, selector: #selector(willResignActive), name: UIApplication.willResignActiveNotification, object: nil)

        _ = checkStatus()
        // mdf by ak 切换外放和请求权限分开 否则会导致首次安装时不能执行切换操作
        let options: AVAudioSession.CategoryOptions = [.defaultToSpeaker, .allowBluetooth]
        try? AVAudioSession.sharedInstance().setCategory(.playAndRecord, options: options)

        // add by ak 取 nsl token
        BFRecordScreenViewModel.getNlsAccessToken { [weak self] token, appkey in
            BFLog(message: "nls appkey is \(appkey), token is \(token)")
            self?.NeoNuiAPPID = appkey
            self?.NeoNuiToken = token
        }

        // MARK: - 录音机初始化

        recorderManager = BFVoiceRecordManager()

        // 录音进度
        recorderManager?.recorderProgrossHandle = { [weak self] progress in
//            BFLog(1, message: "curr:录音进度--\(progress),\(self?.isRecording ?? false)")
            self?.drawProgressIndication(progress: (progress.isNaN || progress.isInfinite) ? 0 : progress)
        }

        // 录音字幕回调
        recorderManager?.subtitleRecordHandle = { [weak self] eventCode, recordId, asrResult, audioFilePath, _ in
            BFLog(1, message: "eventcode:\(eventCode), recordid:\(String(describing: recordId)), asr:\(String(describing: asrResult))")
            
            DispatchQueue.global().async { [weak self] in
                let newSubtitle = PQEditSubTitleModel()
                newSubtitle.recordId = recordId

                if asrResult == nil || (asrResult?.count ?? 0) <= 0 {
                    BFLog(1, message: "onNuiEventCallback 识别结果为空？？？？不能生成字幕数据")
                    return
                }
                // 只有为24时才有字幕返回
                if eventCode != 24 || recordId == nil{
                    BFLog(1, message: "onNuiEventCallback eventCode != 24 || recordId == nil")

                    return
                }
                
                let dicResult: [String: Any]? = jsonStringToDictionary(asrResult!)
                let header = dicResult?["header"] as? [String: Any]
                let payload = dicResult?["payload"] as? [String: Any]
                BFLog(1, message: "onNuiEventCallback event 识别结果:) \(payload?["result"] ?? "") ,taskId:\((header?["task_id"] as? String) ?? "taskId"), 识别时间：\(((payload?["begin_time"]) as? Int) ?? 0) ~ \(((payload?["time"]) as? Int) ?? 0) startTime:\(self?.recorderManager?.voiceModel?.startCMTime.seconds ?? 0.0)")
                // 1，保存字幕数据 begin_time是开始出现文字的时间，time 是结束文字出现的时间 单位都为毫秒，都是相对于录制音频数据整段时间。self.recorderManager.voiceModel?.startCMTime.seconds 为开始的录制的时间，开始和结束都要加上这个时差
                let taskID = "aa"
                // 这里加300ms 是因为返回结果为了切到字，时长提前一些时间，具体时间官方没说和原音频有关系。这里我们先延后300ms 单位：毫秒。
                var tempVoice: PQVoiceModel?
                var tempItem: BFRecordItemModel?
                // 1:先通过titleTaskId来找是否存在录音
                self?.itemModels.enumerated().forEach { index, item in
                    if tempVoice == nil {
                        tempVoice = item.voiceStickers.first { voice in
                            voice.recordId == recordId
                        }
                        if tempVoice != nil {
                            tempItem = item
                        }
                        BFLog(3, message: "字幕回调-找到录音文件:taskID=\(taskID ),audioFilePath=\(audioFilePath ?? ""),index=\(index)")
                    }
                }
                // 2:如果通过titleTaskId没找到录音文件,则通过
                if tempVoice == nil && (self?.isRecording ?? false) && self?.itemModels[self?.currItemModelIndex ?? 0].voiceStickers.last?.recordId == nil {
                    BFLog(3, message: "字幕回调-如果通过titleTaskId没找到录音文件:taskID=\(taskID ),audioFilePath=\(audioFilePath ?? "")")
                    tempVoice = self?.itemModels[self?.currItemModelIndex ?? 0].voiceStickers.last
                }
                // 3:如果通过titleTaskId跟audioFilePath都没找到录音文件,则默认为recorderManager?.voiceModel
                guard let currentVoice = (tempVoice ?? self?.recorderManager?.voiceModel) ?? self?.itemModels[self?.currItemModelIndex ?? 0].voiceStickers.last else {
                    BFLog(3, message: "字幕回调-最终没找到录音文件:taskID=\(taskID ),audioFilePath=\(audioFilePath ?? "")")
                    return
                }
                guard let currentItem = tempItem ?? self?.itemModels[self?.currItemModelIndex ?? 0] else {
                    BFLog(3, message: "字幕回调-最终没找到录音文件:taskID=\(taskID ),audioFilePath=\(audioFilePath ?? "")")
                    return
                }
                newSubtitle.timelineIn = currentVoice.startCMTime + CMTime(seconds: Float64((((payload?["begin_time"]) as? Int) ?? 0) + 300) / 1000.0, preferredTimescale: 1000)
                newSubtitle.timelineOut = currentVoice.startCMTime + CMTime(seconds: Float64(((payload?["time"]) as? Int) ?? 0) / 1000.0, preferredTimescale: 1000)
                newSubtitle.audioFilePath = currentVoice.wavFilePath
                BFLog(3, message: "字幕回调-字幕返回begin:\(((payload?["begin_time"]) as? Int) ?? 0),timelineIn:\(newSubtitle.timelineIn.seconds), end:\(((payload?["time"]) as? Int) ?? 0),timelineOut:\(newSubtitle.timelineOut.seconds)")
                if (newSubtitle.timelineIn - currentVoice.startCMTime).seconds > 0.1 {
                    BFLog(1, message: "卡在录音尾巴上了1")
                    newSubtitle.timelineIn = newSubtitle.timelineIn - CMTime(seconds: 0.1, preferredTimescale: 1000)
                }
                let showText = ((payload?["result"]) as? String) ?? ""
                newSubtitle.text = showText
//                newSubtitle.audioFilePath = audioFilePath ?? ""
                BFLog(1, message: "添加字幕数据 timelineIn \(newSubtitle.timelineIn.seconds) timelineOut \(newSubtitle.timelineOut.seconds) text: \(newSubtitle.text) 音频路径为：\(audioFilePath ?? "bb") 传入的地址：\(currentVoice.wavFilePath ?? "aa")")
                newSubtitle.setting = self?.subtitleSettingView.subtitle.setting ?? BFSubTitileSettingModel()
                currentItem.titleStickers.append(newSubtitle)
            }
        }

        // MARK: -  录音结束
        recorderManager?.endRecordHandle = { [weak self, weak recorderManager] voideModel, _ in
            if let sself = self, let model = voideModel, FileManager.default.fileExists(atPath: model.wavFilePath ?? "") {
                // 加入到语音数组里

                model.endCMTime = sself.currentAssetProgress
                BFLog(1, message: "对应关系：录制结束文件地址：\((model.wavFilePath ?? "").replacingOccurrences(of: documensDirectory, with: "")) 开始录音前地址：\((self?.recorderManager?.voiceModel?.wavFilePath ?? "aa").replacingOccurrences(of: documensDirectory, with: ""))-  \(model.startCMTime.seconds)-\(model.endCMTime.seconds)-\(model.endCMTime.seconds - model.startCMTime.seconds)")
                /// 注：录音机回调的录音时长大于一秒，而业务逻辑计算的会小于一秒
                if (model.endCMTime.seconds - model.startCMTime.seconds) < 1 {
                    // 取消录制
                    sself.recordManagerCancelRecord(voiceModel: model)
                    return
                }
                // ********** 开始处理冲突的录制部分
                let newRange = CMTimeRange(start: model.startCMTime, end: model.endCMTime)

                var deletedVoices = [(PQVoiceModel, Int)]()
                // 要删除的字幕
                var deletedTitlesTemp = [PQEditSubTitleModel]()

                // 查找要删除的音频和字幕数据
                for (i, m) in sself.itemModels[sself.currItemModelIndex].voiceStickers.enumerated() {
                    let originRange = CMTimeRange(start: m.startCMTime, end: CMTime(seconds: m.endCMTime.seconds - 0.02, preferredTimescale: 1000))
                    if CMTimeRangeGetIntersection(originRange, otherRange: newRange).duration.seconds > 0 {
                        deletedVoices.append((m, i))
                        deletedTitlesTemp += sself.deleteTitles(voiceModel: m)
                        continue
                    }
                }
                // 删除冲突的音频
                deletedVoices.forEach { m, _ in
                    sself.itemModels[sself.currItemModelIndex].voiceStickers.removeAll { tempM in
                        m.wavFilePath == tempM.wavFilePath
                    }
                }

                BFLog(1, message: "添加录音文件:\(model.startCMTime.seconds) -- \(model.endCMTime.seconds)")
                sself.itemModels[sself.currItemModelIndex].voiceStickers.append(model)
                // 录制结束回调
                self?.recordEndHandle?(model)
                // 如果是图片素材同时有需要删除的录音时需要调整录音文件开始结束时间
                // warning: 图片录制的时候应该只能在结尾处录制
                if sself.itemModels[sself.currItemModelIndex].mediaType == .Image {
                    if deletedVoices.count > 0 {
                        // 如果是图片先排序在计算区间
                        sself.itemModels[sself.currItemModelIndex].voiceStickers = sself.itemModels[sself.currItemModelIndex].voiceStickers.sorted { voice1, voice2 in
                            voice1.startCMTime.seconds < voice2.startCMTime.seconds
                        }
                        for (index, item) in sself.itemModels[sself.currItemModelIndex].voiceStickers.enumerated() {
                            if index > 0 {
                                // 注：开始时间减去duration or 等一前一段录音的结束时间
                                let tempDuration = item.endCMTime.seconds - item.startCMTime.seconds
                                item.startCMTime = CMTime(seconds: sself.itemModels[sself.currItemModelIndex].voiceStickers[index - 1].endCMTime.seconds, preferredTimescale: 1000)
                                item.endCMTime = CMTime(seconds: item.startCMTime.seconds + tempDuration, preferredTimescale: 1000)
                            }
                            BFLog(message: "录制结束重新排序录音文件：\(index)-\(item.wavFilePath ?? "")-\(item.startCMTime.seconds)-\(item.endCMTime.seconds)-\(item.endCMTime.seconds - item.startCMTime.seconds)")
                        }
                    }
                }
                // ***********处理冲突的录音部分资源 end

                // 添加撤销记录点
                var event = WithDrawModel(type: 2, timestamp: model.startCMTime)
                event.deletedVoices = deletedVoices
                event.deletedTittles = deletedTitlesTemp
                sself.events.append(event)

                if sself.itemModels[sself.currItemModelIndex].mediaType == .Image {
                    var duration: CMTime = .zero
                    sself.itemModels[sself.currItemModelIndex].voiceStickers.forEach { temp in
                        BFLog(1, message: "录制结束-最终：\(temp.wavFilePath ?? "")-\(temp.startCMTime.seconds)-\(temp.endCMTime.seconds)-\(temp.endCMTime.seconds - temp.startCMTime.seconds)")
                        temp.duration = "\(temp.endCMTime.seconds - temp.startCMTime.seconds)"
                        duration = duration + temp.endCMTime - temp.startCMTime
                    }
                    sself.itemModels[sself.currItemModelIndex].materialDuraion = duration
                    self?.isEndPlay = true
                    // 录制结束显示播放按钮
                    sself.playBtn.isSelected = sself.itemModels[sself.currItemModelIndex].voiceStickers.count <= 0
                    sself.playBtn.isHidden = sself.playBtn.isSelected
                }
                DispatchQueue.main.async { [weak self] in
                    // 录音完，重绘撤销按钮，更新录音按钮，
                    self?.changeWithDrawBtnLayout(2)
                    // 注：在录制结束时矫正当前位置，避免跟指针无法对其
                    self?.indirectionView?.resetCurrentItem(start: model.startCMTime.seconds, end: model.endCMTime.seconds)
                    if deletedVoices.count > 0 {
                        /// 重绘录音进度视图
                        self?.resetAllIndirectionView()
                    }
                    // 矫正进度
                    self?.resetCurrentProgress()
                    self?.deleteRecordBtn.isHidden = true
                    self?.recordBtn.isHidden = (self?.itemModels[self?.currItemModelIndex ?? 0].mediaType == .Image && (self?.isEndPlay ?? false)) ? false : (self?.isEndPlay ?? false)
                }
                sself.currentPlayRecordIndex = -3 // 刚录音完，不需要播放录音
                BFLog(3, message: "重置播放index-\(#function) = \(sself.currentPlayRecordIndex)")
                recorderManager?.voiceModel = nil
            } else {
                BFLog(2, message: "数据出错！！！！\(voideModel?.wavFilePath ?? "")")
            }
        }
        recorderManager?.cancelRecordHandle = { [weak self] voiceModel in
            // 取消录制
            self?.recordManagerCancelRecord(voiceModel: voiceModel)
        }

        recorderManager?.NeoNuiDebugHandle = { [weak self] msg,isShow in
            self?.neoNuiDebugLabel.text = "字幕服务:\(msg ?? "")"
            if(isShow){
                cShowHUB(superView: self?.view, msg: msg)
            }
        
        }

        recorderManager?.AudioQueueRecoderDebugHandle = { [weak self] msg,isShow in
            self?.audioQueueRecoderLabel.text = "录音机:\(msg ?? "")"
            if(isShow){
                cShowHUB(superView: self?.view, msg: msg)
            }
        }

        view.backgroundColor = .black
        view.addSubview(collectionView)
        
        fetchVideo()
        
        view.addSubview(bottomeView)
//        view.addSubview(subtitleLabel)
        view.addSubview(playBtn)
        view.addSubview(avatarView)
//        view.addSubview(openCameraBtn)
//        view.addSubview(drawPinBtn)
        view.addSubview(subtitleBtn)
        view.addSubview(soundSettingBtn)
        view.addSubview(subtitleSettingView)

        view.addSubview(audioSettingView)

        bottomeView.addSubview(progreddL)
        //        view.addSubview(toolV)
        bottomeView.addSubview(recordBtn)
        bottomeView.addSubview(deleteRecordBtn)
        bottomeView.addSubview(withDrawBtn)
        //        bottomeView.addSubview(changeVoiceBtn)
        bottomeView.addSubview(progressThumV)

        avatarView.recordEndCallBack = { _, materialsModel in
            BFLog(message: "新录制完成::::\(materialsModel?.locationPath ?? "")")
        }


        // 字幕设置回调
        subtitleSettingView.subtitleSettingCallBack = { [weak self] subtitileModel in

            self?.setSubtitleStyle(settingModel: subtitileModel.setting)
            // mdf by ak 这里是设置字幕开关回调
            if self?.subTitleBtnClickHandle != nil {
                self?.subTitleBtnClickHandle!(subtitileModel.setting.subtitleIsShow)
            }
        }
        // 编辑字幕完成
        subtitleEditView.editSubtitleDone = { [weak self] newtext, index in

            // 更新缓存数据
            if index < (self?.itemModels[self?.currItemModelIndex ?? 0].titleStickers.count ?? 0) {
                if newtext.count == 0 {
                    // 删除数据
                    BFLog(message: "清空字幕操作要删除原字幕 sticker 数据")
                    self?.itemModels[self?.currItemModelIndex ?? 0].titleStickers.remove(at: index)
                } else {
                    // 更新数据
                    self?.itemModels[self?.currItemModelIndex ?? 0].titleStickers[index].text = newtext
                }
            }

            self?.setSubtitleStyle(settingModel: (self?.subtitleSettingView.subtitle.setting)!)
        }

        layoutsubview()

        view.addSubview(subtitleEditView)

        if !SWNetRequest.isNetReachabled() {
            cShowHUB(superView: view, msg: "网络不佳，字幕可能无法生成")
        }
        if PQENVUtil.shared.channel == "Development" {
            view.addSubview(neoNuiDebugLabel)
            view.addSubview(audioQueueRecoderLabel)
        }
    }

    @objc func editSubtitle() {
        BFLog(message: "编辑字幕 index:\(showSubtitleIndex)")
        if !isNormalPlaying, subtitleLabel.text?.count != 0 {
            subtitleEditView.isHidden = false
            subtitleEditView.textView.becomeFirstResponder()
            subtitleEditView.setNewText(text: subtitleLabel.text ?? "", index: showSubtitleIndex)
        } else {
            BFLog(message: "播放状态,内容为空时不可以编辑字幕")
        }
    }

    /// 更新字幕，在回放时使用
    /// - Parameter time: 当前播放的进度
    func updateSubtitle(time: CMTime) {
        BFLog(message: "currTime is \(CMTimeGetSeconds(time))")
        if isRecording || !subtitleSettingView.subtitle.setting.subtitleIsShow || currentPlayRecordIndex == -3 {
            // currentPlayRecordIndex == -3 代表刚录完音，不要找字幕，会因为小幅度回退seek导致上一个字幕出现
            if subtitleLabel.text?.count ?? 0 > 0 {
                subtitleLabel.text = ""
                subtitleLabel.backgroundColor = UIColor.clear
            }
            return
        }

        if showSubtitleIndex > -1, showSubtitleIndex < itemModels[currItemModelIndex].titleStickers.count {
            let currShowSubtitle = itemModels[currItemModelIndex].titleStickers[showSubtitleIndex]

            if CMTimeCompare(currShowSubtitle.timelineIn, time) <= 0, CMTimeCompare(currShowSubtitle.timelineOut, time) > 0 {
                // 已存在字幕
                if subtitleLabel.text != currShowSubtitle.text {
                    subtitleLabel.text = currShowSubtitle.text
                    setSubtitleStyle(settingModel: subtitleSettingView.subtitle.setting)
                }
                return
            }
            showSubtitleIndex = -1
        }

        var findShowSubtitle: PQEditSubTitleModel?
        for (index, subtitle) in itemModels[currItemModelIndex].titleStickers.enumerated() {
            if CMTimeCompare(subtitle.timelineIn, time) <= 0, CMTimeCompare(subtitle.timelineOut, time) > 0, subtitle.audioFilePath.count > 0 { //  audioFilePath.count 这个条件是确保这个字幕有对应录音
                findShowSubtitle = subtitle
//                BFLog(1, message: "找到要显示的字幕 in \((findShowSubtitle?.timelineIn ?? .zero).seconds) out \((findShowSubtitle?.timelineOut ?? .zero).seconds) text:\(findShowSubtitle?.text ?? "") currTime is \(CMTimeGetSeconds(time))")
                showSubtitleIndex = index
                break
            }
        }

        if findShowSubtitle != nil, subtitleLabel.text != findShowSubtitle!.text {
            subtitleLabel.text = findShowSubtitle!.text
            setSubtitleStyle(settingModel: subtitleSettingView.subtitle.setting)

        } else {
            if subtitleLabel.text?.count ?? 0 > 0 {
                subtitleLabel.text = ""
                subtitleLabel.backgroundColor = UIColor.clear
            }
        }
    }

    /// 设置字幕样式和位置
    /// - Parameter settingModel: 样式model
    func setSubtitleStyle(settingModel: BFSubTitileSettingModel) {
        subtitleBtn.setImage(imageInRecordScreenKit(by: settingModel.subtitleIsShow ? "subtitleBtn_on" : "subtitleBtn_off"), for: .normal)
        if settingModel.subtitleIsShow {
            // 设置样式
            subtitleLabel.strokeColor = settingModel.strokeColor
            subtitleLabel.strokeWidth = settingModel.strokeWidth * cScreenWidth / 1080
            subtitleLabel.font = UIFont.boldSystemFont(ofSize: CGFloat(settingModel.subtitleSize) * cScreenWidth / 1080)
            subtitleLabel.textColor = settingModel.fontColor

            let leftPoint: CGFloat = 37

            let height = sizeTextFits(attributedText: nil, text: subtitleLabel.text ?? "", numberOfLines: 0, font: subtitleLabel.font, maxSize: CGSize(width: cScreenWidth * 0.75, height: CGFloat.greatestFiniteMagnitude)).height

            subtitleLabel.backgroundColor = subtitleSettingView.subtitle.setting.backgroundColor
            // 下 //设置位置
            if subtitleSettingView.subtitle.setting.subtitlePoint == 0 {
                subtitleLabel.frame = CGRect(x: leftPoint, y: cScreenHeigth * 0.70 - height, width: cScreenWidth - 37 * 2, height: height)
            } else if subtitleSettingView.subtitle.setting.subtitlePoint == 1 { // 中
                subtitleLabel.frame = CGRect(x: leftPoint, y: cScreenHeigth / 2.0 - height / 2.0, width: cScreenWidth - 37 * 2, height: height)
            } else { // 上
                subtitleLabel.frame = CGRect(x: leftPoint, y: cScreenHeigth * 0.12, width: cScreenWidth - 37 * 2, height: height)
            }
            // 打开的时候主动调用一次显示字幕
            updateSubtitle(time: currentAssetProgress)
        } else {
            subtitleLabel.text = ""
        }
    }

    func layoutsubview() {
        bottomeView.snp.makeConstraints { make in
            make.left.bottom.right.equalToSuperview()
            make.height.equalTo(adapterWidth(width: 180))
        }

        progreddL.snp.makeConstraints { make in
            make.width.equalTo(100)
            make.centerX.equalToSuperview()
            make.top.equalToSuperview().offset(-5)
            make.height.equalTo(18)
        }

        withDrawBtn.snp.makeConstraints { make in
            make.left.equalToSuperview()
            make.width.height.equalTo(65)
            make.top.equalTo(88)
        }
        //        changeVoiceBtn.snp.makeConstraints { make in
        //            make.right.equalToSuperview()
        //            make.top.width.height.equalTo(withDrawBtn)
        //        }

        recordBtn.snp.makeConstraints { make in
            make.left.equalTo(withDrawBtn.snp.right)
            make.right.equalTo(-65)
            make.height.equalTo(42)
            make.top.equalTo(withDrawBtn).offset(6)
        }
        deleteRecordBtn.snp.makeConstraints { make in
            make.left.equalTo(withDrawBtn.snp.right)
            make.right.equalTo(-65)
            make.height.equalTo(42)
            make.top.equalTo(withDrawBtn).offset(6)
        }

        //        openCameraBtn.snp.makeConstraints { make in
        //            make.right.equalToSuperview().offset(-12)
        //            make.top.equalToSuperview().offset(98)
        //            make.width.equalTo(80)
        //            make.height.equalTo(124)
        //        }
        //
        //        drawPinBtn.snp.makeConstraints { make in
        //            make.right.equalTo(openCameraBtn)
        //            make.top.equalTo(openCameraBtn.snp.bottom).offset(18)
        //            make.width.equalTo(80)
        //            make.height.equalTo(124)
        //        }

        soundSettingBtn.snp.makeConstraints { make in
            make.right.equalToSuperview().offset(-12)
            make.top.equalToSuperview().offset(98)
            make.width.equalTo(40)
            make.height.equalTo(62)
        }

        subtitleBtn.snp.makeConstraints { make in
            make.right.equalTo(soundSettingBtn)
            make.top.equalTo(soundSettingBtn.snp.bottom).offset(18)
            make.width.equalTo(40)
            make.height.equalTo(62)
        }

        withDrawBtn.imageEdgeInsets = UIEdgeInsets(top: -withDrawBtn.imageView!.height, left: 0, bottom: 0, right: -withDrawBtn.titleLabel!.width)
        withDrawBtn.titleEdgeInsets = UIEdgeInsets(top: withDrawBtn.titleLabel!.height + 2, left: -withDrawBtn.imageView!.width, bottom: 0, right: 0)

        changeVoiceBtn.imageEdgeInsets = UIEdgeInsets(top: -changeVoiceBtn.imageView!.height - 2, left: 0, bottom: 0, right: -changeVoiceBtn.titleLabel!.width)
        changeVoiceBtn.titleEdgeInsets = UIEdgeInsets(top: changeVoiceBtn.titleLabel!.height + 2, left: -changeVoiceBtn.imageView!.width, bottom: 0, right: 0)
    }

    //    public override func viewWillLayoutSubviews() {
    //        super.viewWillLayoutSubviews()
    //
    //    }

    // MARK: - 按钮事件响应

    override public func backBtnClick() {
        pause()
    }

    // 触发拖曳手势后，执行的动作
    @objc func pan(recognizer: UIPanGestureRecognizer) {
        // 设置 UIView 新的位置
        if !checkStatus(show: false) {
            return
        }
        switch recognizer.state {
        case .began:
            touchStart = recognizer.location(in: bottomeView)
            beginOnStartBtn = (beginOnStartBtn || recordBtn.frame.contains(touchStart))
            BFLog(1, message: "beginOnStartBtn:\(beginOnStartBtn)")
        case .changed:
            if beginOnStartBtn == true {
                let nowPoint = recognizer.location(in: bottomeView)
                BFLog(1, message: "nowPoint x: \(nowPoint.x) y:\(nowPoint.y)")
                if recordBtn.frame.contains(nowPoint) {
                } else {}
            }

        case .ended:
            BFLog(1, message: "移动结束 \(beginOnStartBtn)")
            if beginOnStartBtn == true {
                beginOnStartBtn = false
                endRecord()
                //                if recordBtn.frame.contains(point) {
                //                    // 结束录制
                //                } else {
                //                    cancleRecord()
                //                }
            }

        default:
            BFLog(1, message: "pan default")
        }
    }

    @objc func openCamera() {
        if avatarView.isHidden {
            avatarView.isHidden = false
            avatarView.openCamera()
        } else {
            avatarView.isHidden = true
            avatarView.closeCamera()
        }
    }

    @objc func drawPin() {}

    // 字幕
    @objc func subTitleClick() {
        BFLog(message: "subTitle Click ")
        subtitleSettingView.isHidden = !subtitleSettingView.isHidden

        updateSubtitle(time: currentAssetProgress)
    }

    // 声音设置
    @objc func soundSetting() {
        BFLog(message: "设置声音")

        audioSettingView.isHidden = false
    }

    /// 删除指定段落的所有字幕 数据
    /// - Parameter voiceModel: 删除的音频数据
    func deleteTitles(voiceModel: PQVoiceModel) -> [PQEditSubTitleModel] {
        BFLog(message: "itemModels[currItemModelIndex].titleStickers  删除前：\(itemModels[currItemModelIndex].titleStickers.count) model.startCMTime.seconds: \(voiceModel.startCMTime.seconds) model.end: \(voiceModel.endCMTime.seconds)")
        var deleteTemp = [PQEditSubTitleModel]()

        // 从原数组中删除
        itemModels[currItemModelIndex].titleStickers.removeAll { m in
            let su = (m.recordId == voiceModel.recordId)
            if su {
                deleteTemp.append(m)
            }
            return su
        }

        BFLog(message: "itemModels[currItemModelIndex].titleStickers  删除后：\(itemModels[currItemModelIndex].titleStickers.count)")

        // 清空字幕UI
        subtitleLabel.text = ""
        subtitleLabel.backgroundColor = UIColor.clear

        return deleteTemp
    }

    @objc func deleteRecordAction() {
        if !isDragingProgressSlder, isStopAtRecordRange != -1, isStopAtRecordRange < itemModels[currItemModelIndex].voiceStickers.count {
            let model = itemModels[currItemModelIndex].voiceStickers[isStopAtRecordRange]
            itemModels[currItemModelIndex].voiceStickers.remove(at: isStopAtRecordRange)
            indirectionView?.deleteItem(index: isStopAtRecordRange)
            var event = WithDrawModel(type: 3, timestamp: currentAssetProgress, deletedVoices: [(model, isStopAtRecordRange)])
            event.deletedTittles = deleteTitles(voiceModel: model)
            events.append(event)

            // 注：删除录音后图片素材需要回撤指针进度,同时后面录音往前挪
            if itemModels[currItemModelIndex].mediaType == .Image {
                let currDuration = model.endCMTime - model.startCMTime
                itemModels[currItemModelIndex].materialDuraion = itemModels[currItemModelIndex].materialDuraion - currDuration
                currentAssetProgress = model.startCMTime
                // 更新进度
                resetCurrentProgress()
                for (index, item) in itemModels[currItemModelIndex].voiceStickers.enumerated() {
                    if index >= isStopAtRecordRange {
                        // 注：开始时间减去duration or 等一前一段录音的结束时间
                        item.startCMTime = item.startCMTime - currDuration
                        item.endCMTime = item.endCMTime - currDuration

                        let titlsList = itemModels[currItemModelIndex].titleStickers.filter { tm in
                            tm.recordId == item.recordId
                        }

                        for titleM in titlsList {
                            titleM.timelineIn = titleM.timelineIn - currDuration
                            titleM.timelineOut = titleM.timelineOut - currDuration
                        }
                    }
                }
                /// 重绘录音进度视图
                resetAllIndirectionView()

                // 判断是否无录音了
                if itemModels[currItemModelIndex].materialDuraion == .zero {
                    playBtn.isSelected = true
                    playBtn.isHidden = playBtn.isSelected
                }
            }
            searchStopAtRecordRange()
            changeWithDrawBtnLayout(3)
        }
    }

    @objc func startRecord() {
        //        BFLog(1, message: "start \(UIControl.Event.touchDown)")
        if !checkStatus(show: true) {
            BFLog(message: "没有录音权限不能录制")
            return
        }

        // 停止进度条滚动
        if isDragingProgressSlder {
            return
        }

//        progressThumV.progressView.isUserInteractionEnabled = false
//        collectionView.isScrollEnabled = false
        // 开始录制时清除屏幕上的字幕
        subtitleLabel.text = ""
        subtitleLabel.backgroundColor = UIColor.clear

        pause()
        if recorderManager == nil {
            BFLog(message: "录音机初始化错误！！！")
            return
        }

        recorderManager?.cancelTitleService()
        // 开始时间
        beginOnStartBtn = true
        isRecording = true

        UIApplication.shared.keyWindow?.isUserInteractionEnabled = false

        let model = PQVoiceModel()
        model.startCMTime = currentAssetProgress
        model.currIndex = currItemModelIndex
//        model.recordId = getUniqueId(desc: "recordId")
        model.volume = 100
        recorderManager?.voiceModel = model
        BFLog(3, message: "开始录制-开始：currentAssetProgress=\(currentAssetProgress.seconds),cuInde=\(currItemModelIndex),\(model)")
        recorderManager?.startRecord()
        recorderManager?.audioRecorder?.startNeoNui(NeoNuiToken ?? "", appid: NeoNuiAPPID ?? "")

        if !avatarView.isHidden {
            avatarView.beginRecord()
        }

        rscurrentManager.startRecord()
        // 录制中不显示播放按钮
        playBtn.isSelected = true
        playBtn.isHidden = playBtn.isSelected
        let cell = collectionView.cellForItem(at: IndexPath(item: currItemModelIndex, section: 0)) as? BFImageCoverViewCell
        if cell?.contentView.viewWithTag(100_100)?.superview == nil {
            cell?.contentView.addSubview(subtitleLabel)
        }
    }

    @objc func endRecord() {
        pause()
        // 存储录音
        if !isRecording {
            return
        }
        beginOnStartBtn = false

        BFLog(1, message: "停止录音-")
        recordBtn.isEnabled = false
        DispatchQueue.main.asyncAfter(deadline: .now() + 0.25) { [weak self] in
            self?.recordBtn.isEnabled = true
        }

        isRecording = false
//        progressThumV.progressView.isUserInteractionEnabled = true
//        collectionView.isScrollEnabled = true

        UIApplication.shared.keyWindow?.isUserInteractionEnabled = true
        recorderManager?.stopRecord(isCancel: false)

        if currentAssetProgress.seconds - (recorderManager?.voiceModel?.startCMTime.seconds ?? 0) >= 1.0 {
        } else {
            changeProgress(changCMTime: recorderManager?.voiceModel?.startCMTime ?? .zero)
        }

        /*
         if currentAssetProgress.seconds - (recorderManager?.voiceModel?.startCMTime.seconds ?? 0 ) >= 1.0 {
             recorderManager?.stopRecord(isCancel: false)
             changeWithDrawBtnLayout(2)
         }else{
             recorderManager?.stopRecord(isCancel: true)
             changeProgress(progress: Float(recorderManager?.voiceModel?.startCMTime.seconds ?? 0))
         }
         */
        if !avatarView.isHidden {
            avatarView.endRecord()
        }
    }

    @objc func cancleRecord() {
        isRecording = false
        recorderManager?.stopRecord(isCancel: true)
//        progressThumV.progressView.isUserInteractionEnabled = true
//        collectionView.isScrollEnabled = true
        UIApplication.shared.keyWindow?.isUserInteractionEnabled = true
        pause()
    }

    /// 不足一秒，主动取消录制
    /// - Parameter voiceModel: <#voiceModel description#>
    @objc func recordManagerCancelRecord(voiceModel: PQVoiceModel?) {
        // 录制结束回调
        recordEndHandle?(voiceModel)
        BFLog(3, message: "开始录制-取消：currentAssetProgress=\(currentAssetProgress.seconds),cuInde=\(currItemModelIndex),currIndex=\(voiceModel?.currIndex ?? 0),\(String(describing: voiceModel)),\(String(describing: recorderManager?.voiceModel))")
        if voiceModel != nil, currentAssetProgress.seconds - (recorderManager?.voiceModel?.startCMTime.seconds ?? 0) < 1.0 {
            cShowHUB(superView: nil, msg: "最短录制1秒")
            recorderManager?.cancelTitleService()
        }
        // 删除文件
        recorderManager?.deleteFile(outfile: voiceModel?.wavFilePath ?? "")
        var subtitleCount = itemModels[currItemModelIndex].titleStickers.count
        BFLog(2, message: "删除\(voiceModel?.wavFilePath ?? "")对应的字幕 前 count\(subtitleCount)")
        if subtitleCount > 0 {
            for title in itemModels[currItemModelIndex].titleStickers {
                if title.recordId == voiceModel?.recordId ?? "none" {
                    if let index = itemModels[currItemModelIndex].titleStickers.firstIndex(of: title) {
                        itemModels[currItemModelIndex].titleStickers.remove(at: index)
                    }
                }
            }
        }
        subtitleCount = itemModels[currItemModelIndex].titleStickers.count
        BFLog(2, message: "删除\(voiceModel?.wavFilePath ?? "")对应的字幕  后 count\(subtitleCount)")
        /// 重置进度
        if voiceModel?.currIndex == currItemModelIndex {
            if itemModels[currItemModelIndex].mediaType == .Image {
                currentAssetProgress = CMTime(seconds: recorderManager?.voiceModel?.startCMTime.seconds ?? 0, preferredTimescale: 1000)
                if currentAssetProgress.seconds >= itemModels[currItemModelIndex].materialDuraion.seconds {
                    currentAssetProgress = CMTime(seconds: itemModels[currItemModelIndex].materialDuraion.seconds, preferredTimescale: 1000)
                }
                resetCurrentProgress()
            }
            // 移除
            resetAllIndirectionView()
        }
        recorderManager?.voiceModel = nil
    }

    // 撤销
    @objc func withdrawAction() {
        pause()
        if let action = events.last {
            let jumpTime = action.timestamp
            if action.type == 2 {
                // 撤销录制
                if let modelIndex = itemModels[currItemModelIndex].voiceStickers.firstIndex(where: { mod in
                    CMTimeCompare(mod.startCMTime, action.timestamp) == 0
                }) {
                    // 移除音频
                    let model = itemModels[currItemModelIndex].voiceStickers[modelIndex]
                    itemModels[currItemModelIndex].voiceStickers.remove(at: modelIndex)
                    indirectionView?.deleteItem(index: modelIndex)

                    // 删除对应字幕
                    itemModels[currItemModelIndex].titleStickers.removeAll { m in
                        m.audioFilePath == model.wavFilePath
                    }

                    // 恢复被覆盖的音频
                    var tuples = action.deletedVoices
                    if tuples != nil, tuples!.count > 0 {
                        tuples!.sort { tuple1, tuple2 in
                            tuple1.1 < tuple2.1
                        }
                        tuples?.forEach { tuple in
                            itemModels[currItemModelIndex].voiceStickers.insert(tuple.0, at: tuple.1)
                        }
                    }
                    // 恢复被覆盖的字幕
                    let titleTuples = action.deletedTittles
                    if titleTuples != nil, titleTuples!.count > 0 {
                        titleTuples?.forEach { titleTuple in
                            itemModels[currItemModelIndex].titleStickers.append(titleTuple)
                        }
                    }
//                    jumpTime = model.startCMTime.seconds

                    if itemModels[currItemModelIndex].mediaType == .Image {
                        itemModels[currItemModelIndex].materialDuraion = model.startCMTime
                    }
                }
            } else if action.type == 3 {
                // 恢复删除录音
                var tuples = action.deletedVoices
                tuples?.sort { tuple1, tuple2 in
                    tuple1.1 < tuple2.1
                }
                if itemModels[currItemModelIndex].mediaType == .Video {
                    tuples?.forEach { tuple in
                        itemModels[currItemModelIndex].voiceStickers.insert(tuple.0, at: tuple.1)
                    }
                } else {
                    tuples?.forEach { tuple in
                        itemModels[currItemModelIndex].voiceStickers.insert(tuple.0, at: tuple.1)
                        let currDuration = tuple.0.endCMTime - tuple.0.startCMTime
                        itemModels[currItemModelIndex].materialDuraion = itemModels[currItemModelIndex].materialDuraion + currDuration
                        for (index, item) in itemModels[currItemModelIndex].voiceStickers.enumerated() {
                            if index > tuple.1 {
                                // 注：之后的录音和字幕增加对应的时长
                                item.startCMTime = item.startCMTime + currDuration
                                item.endCMTime = item.endCMTime + currDuration

                                let titlsList = itemModels[currItemModelIndex].titleStickers.filter { tm in
                                    tm.recordId == item.recordId
                                }

                                for titleM in titlsList {
                                    titleM.timelineIn = titleM.timelineIn + currDuration
                                    titleM.timelineOut = titleM.timelineOut + currDuration
                                }
                            }
                        }
                    }
                }
                // 恢复字幕
                let titleTuples = action.deletedTittles
                if titleTuples != nil, titleTuples!.count > 0 {
                    titleTuples?.forEach { titleTuple in
                        itemModels[currItemModelIndex].titleStickers.append(titleTuple)
                    }
                }
                if itemModels[currItemModelIndex].mediaType == .Image {
                    itemModels[currItemModelIndex].materialDuraion = itemModels[currItemModelIndex].voiceStickers.last?.endCMTime ?? .zero
                }

            } else {}
            events.removeLast()

            let dur = itemModels[currItemModelIndex].materialDuraion.seconds
            if dur > 0 {
                if itemModels[currItemModelIndex].mediaType == .Image {
                    changeProgress(isBack: true, progress: Float(jumpTime.seconds))
                } else {
                    changeProgress(changCMTime: jumpTime)
                }
            } else {
                changeProgress(progress: Float(0))
            }
            isDragingProgressSlder = false
            currentPlayRecordIndex = -1
            BFLog(3, message: "重置播放index-\(#function) = \(currentPlayRecordIndex)")
            hadPrepareToPlayRecord = false
            progressThumV.progress = jumpTime.seconds

            if let event = events.last {
                changeWithDrawBtnLayout(event.type)
            } else {
                changeWithDrawBtnLayout(0)
                withDrawBtn.isEnabled = false
            }
            searchStopAtRecordRange()
            let itemModel = itemModels[currItemModelIndex]
            /// 重绘录音进度视图
            resetAllIndirectionView()
            // 如果是图片需重置播放按钮
            if itemModel.mediaType == .Image {
                playBtn.isSelected = itemModels[currItemModelIndex].voiceStickers.count <= 0
                playBtn.isHidden = playBtn.isSelected
            }
        }
    }

    @objc func changeVoiceAction() {
        //        nextActionHandle?()
        pause()
    }
    
    @objc func playbtnDown(btn:UIButton){
        let vv = UIView(frame: CGRect(x: 0, y: 0, width: 0.5, height: 0.5))
        vv.backgroundColor = UIColor.init(white: 1, alpha: 0.3)
        btn.setImage(vv.graphicsGetImage(), for: .normal)
    }
    @objc func playbtnCancle(btn:UIButton){
        btn.setImage(imageInRecordScreenKit(by: "preview_play"), for: .normal)
    }

    @objc func playVideo(btn: UIButton) {
        btn.setImage(imageInRecordScreenKit(by: "preview_play"), for: .normal)

        if itemModels[currItemModelIndex].mediaType == .Image && itemModels[currItemModelIndex].voiceStickers.count <= 0 {
            BFLog(message: "图片没有录音无法播放")
            btn.isSelected = true
            return
        }
        btn.isSelected = !btn.isSelected
        if !btn.isSelected {
            pause()
            searchStopAtRecordRange()
        } else {
            events.append(WithDrawModel(type: 1, timestamp: currentAssetProgress))
            changeWithDrawBtnLayout(0)
            play()
        }
    }

    @objc func sliderTouchBegan(sender _: UISlider) {
        isDragingProgressSlder = true
        pause()
    }

    @objc func sliderTouchEnded(sender: UISlider) {
        changeProgress(progress: sender.value)
        isDragingProgressSlder = false
        currentPlayRecordIndex = -1
        BFLog(3, message: "重置播放index-\(#function) = \(currentPlayRecordIndex),\(sender.value)")
        hadPrepareToPlayRecord = false
    }

    @objc func sliderValueDidChanged(sender: UISlider) {
        changeProgress(progress: sender.value)
    }

    func thumbViewEnded(progress: Float) {
        changeProgress(progress: progress)
        isDragingProgressSlder = false
        currentPlayRecordIndex = -1
        hadPrepareToPlayRecord = false
        BFLog(1, message: "isDragingProgressSlder : \(isDragingProgressSlder)")
        BFLog(3, message: "重置播放index-\(#function) = \(currentPlayRecordIndex)")
        searchStopAtRecordRange()
        changeWithDrawBtnLayout(0)
        isEndPlay = (progress == 1)
        recorderManager?.voiceModel = nil
        
        if CMTimeCompare(currentAssetProgress, itemModels[currItemModelIndex].materialDuraion) >= 0  {
            // 视频拖动到最后隐藏录制按钮
            if itemModels[currItemModelIndex].mediaType == .Video  {
                recordBtn.isHidden = true
            } else {
                recordBtn.isHidden = false
                deleteRecordBtn.isHidden = true
                recordBtn.alpha = 1
            }
        }
    }

    // 是否吸附在录音首尾处
    func searchStopAtRecordRange(needAdsorb: Bool = false) {
        // TODO: 滑动，播放暂停，撤销时，判断是否停止录音区间，是则删除相关录音，画笔，头像，字幕
        let elems = itemModels[currItemModelIndex].voiceStickers.enumerated().filter { elem in
            CMTimeCompare(elem.1.startCMTime, currentAssetProgress) <= 0 && elem.1.endCMTime > currentAssetProgress
//            elem.1.startCMTime.seconds <= self.currentAssetProgress.seconds && (elem.1.endCMTime.seconds - 0.001) > self.currentAssetProgress.seconds
        }

        isEndPlay = false

        if elems.count > 0 {
            //  TODO: 停在了录音区间,显示删除按钮
            if needAdsorb {
                if fabs((elems[0].1.endCMTime - currentAssetProgress).seconds) < 0.5 {
                    BFLog(1, message: "吸附在录音结尾, \(elems[0].1.endCMTime.seconds)")
                    //                changeWithDrawBtnLayout(false)
                    changeProgress(changCMTime: elems[0].1.endCMTime)
                    progressThumV.progress = elems[0].1.endCMTime.seconds

//                    deleteRecordBtn.isHidden = true
//                    recordBtn.isHidden = false
//                    isStopAtRecordRange = -1
//                    BFLog(1, message: "停在了录音区间外 \(isStopAtRecordRange)")
                    searchStopAtRecordRange(needAdsorb: false)
                    return
                } else {
                    if fabs((elems[0].1.startCMTime - currentAssetProgress).seconds) < 0.5 {
                        BFLog(1, message: "吸附在录音开始")
                        //                    changeWithDrawBtnLayout(true)
                        changeProgress(changCMTime: elems[0].1.startCMTime)
                        progressThumV.progress = elems[0].1.startCMTime.seconds
                    }
                    deleteRecordBtn.isHidden = false
                    recordBtn.isHidden = true
                    // add by ak 这里 isHidden 无效
                    recordBtn.alpha = 0
                    isStopAtRecordRange = elems.first!.0

                    BFLog(1, message: "停在了录音区间里 \(isStopAtRecordRange), currTime:\(currentAssetProgress.seconds), 录音范围：\(elems[0].1.startCMTime.seconds) - \(elems[0].1.endCMTime.seconds)")
                }
            } else {
                deleteRecordBtn.isHidden = false
                recordBtn.isHidden = true
                recordBtn.alpha = 0

                isStopAtRecordRange = elems.first!.0
                BFLog(1, message: "停在了录音区间里 \(isStopAtRecordRange), currTime:\(currentAssetProgress.seconds), 录音范围：\(elems[0].1.startCMTime.seconds) - \(elems[0].1.endCMTime.seconds)")
            }

        } else {
            deleteRecordBtn.isHidden = true
            recordBtn.isHidden = false
            recordBtn.alpha = 1

            isStopAtRecordRange = -1
            BFLog(1, message: "停在了录音区间外 \(isStopAtRecordRange)")
        }
        pauseTime = currentAssetProgress.seconds
    }

    // MARK: - 权限申请

    func checkStatus(show: Bool = true) -> Bool {
        let status = AVCaptureDevice.authorizationStatus(for: .audio)
        switch status {
        case .denied, .restricted:

            BFLog(message: "麦克风权限被拒绝，请在设置中打开")
            if show {
                let remindData = BFBaseModel()
                remindData.title = "需要开启「麦克风权限」才可以进行录制"
                remindData.summary = ""
                let remindView = BFRemindView(frame: CGRect(x: 0, y: 0, width: cScreenWidth, height: cScreenHeigth))
                remindView.isBanned = true
                remindView.confirmBtn.setTitle("去开启", for: .normal)
                remindView.confirmBtn.setTitleColor(UIColor.hexColor(hexadecimal: "#389AFF"), for: .normal)
                UIApplication.shared.keyWindow?.addSubview(remindView)
                remindView.remindData = remindData
                remindView.cancelBtn.setTitleColor(UIColor.hexColor(hexadecimal: "#389AFF"), for: .normal)
                remindView.remindBlock = { item, _ in
                    if item.tag == 2 {
                        openAppSetting()
                    }
                }
            }
            return false
        case .authorized:
            return true
        case .notDetermined:
            if show {
                requestMicroPhoneAuth()
            }
            return false
        default:
            break
        }

        return false
    }

    func requestMicroPhoneAuth() {
        AVCaptureDevice.requestAccess(for: .audio) { granted in
            if !granted {
                BFLog(message: "麦克风权限被拒绝，请在设置中打开")
            }
        }
    }

    // MARK: - 音视频处理

    func playRecord(at currentT: CMTime, periodicTimeObserver: @escaping (_ time: CMTime, _ currentItem: AVPlayerItem) -> Void, didPlayToEndTime: @escaping (_ recordedInfo: (Int, PQVoiceModel)?, _ currentItem: AVPlayerItem?) -> Void, playFailed _: @escaping (_ recordedInfo: (Int, PQVoiceModel)?, _ currentItem: AVPlayerItem?) -> Void) {
//        if currentPlayRecordIndex >= 0 {
//            if assetPlayer.volume != haveSpeakVolume{
//                assetPlayer.volume = haveSpeakVolume
//            }
//        }else{
//            if assetPlayer.volume != noSpeakVolume {
//                assetPlayer.volume = noSpeakVolume
//            }
//        }

        if itemModels[currItemModelIndex].voiceStickers.first(where: { m in
            CMTimeCompare(m.startCMTime, currentT) <= 0 && CMTimeCompare(currentT, m.endCMTime) <= 0
        }) != nil {
            if assetPlayer.volume != haveSpeakVolume {
                assetPlayer.volume = haveSpeakVolume
            }
        } else {
            if assetPlayer.volume != noSpeakVolume {
                assetPlayer.volume = noSpeakVolume
            }
        }
//        BFLog(1, message: "volume:\(assetPlayer.volume ?? -1)")

        if currentPlayRecordIndex == -3 { // 刚录音完，不需要播放
            return
        }

        // 先排序，再查找下一个需要播放的录音
        let list = itemModels[currItemModelIndex].voiceStickers.sorted { m1, m2 in
            CMTimeCompare(m1.startCMTime, m2.startCMTime) < 0
        }
        let (shouldPlayRecordIndex, recordedAudio) = list.enumerated().first { model in
            CMTimeCompare(model.1.endCMTime, currentT) > 0
        } ?? (-1, nil)

        // 没找到，代表后边没有录音需求了
        guard let recordedAudio = recordedAudio, recordedAudio.wavFilePath.count > 0 else {
            BFLog(3, message: "未找到可播放录音")
            return
        }
//        BFLog(1, message: "当前时间：\(CMTimeGetSeconds(currentT)), 找到的音频：\(recordedAudio.startCMTime.seconds) ～ \(recordedAudio.endCMTime.seconds), \(recordedAudio.wavFilePath ?? "")")

        // 创建播放器
        if (recordPlayer.currentItem?.asset as? AVURLAsset)?.url.lastPathComponent != (recordedAudio.wavFilePath as NSString?)?.lastPathComponent {
            let newItem = AVPlayerItem(url: URL(fileURLWithPath: recordedAudio.wavFilePath))
            BFLog(1, message: "录音播放器初始化：\("replace item")")

            recordPlayer.pause()
            if let playItem = recordPlayer.currentItem {
                NotificationCenter.default.removeObserver(self, name: .AVPlayerItemDidPlayToEndTime, object: playItem)
                recordPlayer.replaceCurrentItem(with: newItem)
            }
            
            recordPlayer.volume = 1
            currentPlayRecordIndex = -1
            hadPrepareToPlayRecord = false
            BFLog(1, message: "录音播放器初始化(有时候不准)")
            BFLog(3, message: "重置播放index-\(#function) = \(currentPlayRecordIndex)")
            NotificationCenter.default.addObserver(forName: .AVPlayerItemDidPlayToEndTime, object: newItem, queue: .main) { [weak self] _ in
                guard let sself = self else {
                    BFLog(3, message: "sself为空AVPlayerItemDidPlayToEndTime")
                    return
                }
                sself.hadPrepareToPlayRecord = false
                sself.currentPlayRecordIndex = -1
//                sself.recordPlayer.volume = 0
//                sself.assetPlayer.volume = sself.noSpeakVolume
                BFLog(3, message: "播放结束")
                didPlayToEndTime((shouldPlayRecordIndex, recordedAudio), newItem)
            }
            recordPlayerTimeObserver?.invalidate()
            recordPlayerTimeObserver = recordPlayer.addPeriodicTimeObserver(forInterval: CMTime(value: 1, timescale: 1000), queue: DispatchQueue.global()) { [weak self, weak recordPlayer] time in
                guard let sself = self, let rPlay = recordPlayer else {
                    BFLog(3, message: "sself为空")
                    return
                }
                BFLog(3, message: "当前播放---\(time),\(time.seconds),\(rPlay.currentItem?.currentTime().seconds ?? 0),\(rPlay.currentItem?.duration.seconds ?? 0)")
                if CMTimeGetSeconds(sself.currenStartPlayTime) <= 0 {
                    BFLog(message: "重新更新开始播放进度\(#function)-\(sself.currenStartPlayTime.seconds)")
                    sself.currenStartPlayTime = time
                }
                let progressTime = CMTime(seconds: CMTimeGetSeconds(time) - CMTimeGetSeconds(sself.currenStartPlayTime), preferredTimescale: 1000)
                BFLog(message: "progressTime进度\(#function)-\(progressTime.seconds)")
                periodicTimeObserver(progressTime, newItem)
//                if (rPlay.currentItem?.currentTime().seconds ?? 0) > (rPlay.currentItem?.duration.seconds ?? 0) - 0.1 {
//                    didPlayToEndTime((shouldPlayRecordIndex, recordedAudio), newItem)
//                }
            } as? NSKeyValueObservation
        }
        if itemModels[currItemModelIndex].mediaType == .Video {
            videoMaterialRecordPlay(at: currentT, shouldPlayRecordIndex: shouldPlayRecordIndex, recordedAudio: recordedAudio)
        } else {
            imageMaterialRecordPlay(at: currentT, shouldPlayRecordIndex: shouldPlayRecordIndex, recordedAudio: recordedAudio)
        }
        BFLog(1, message: "应当播放：\(shouldPlayRecordIndex), 当前播放：\(currentPlayRecordIndex)")
    }

    /// 视频素材录音的播放
    /// - Parameters:
    ///   - currentT: <#currentT description#>
    ///   - shouldPlayRecordIndex: <#shouldPlayRecordIndex description#>
    ///   - recordedAudio: <#recordedAudio description#>
    func videoMaterialRecordPlay(at currentT: CMTime, shouldPlayRecordIndex: Int, recordedAudio: PQVoiceModel) {
        //        if recordPlayer.currentItem?.duration.timescale == 0 {
        //            BFLog(1, message: "时间timescale  == 0")
        //            playFailed((shouldPlayRecordIndex, recordedAudio) as? (Int, PQVoiceModel), recordPlayer.currentItem)
        //            return
        //        }
        synced(currentPlayRecordIndex) { [weak self] in
            guard let sself = self else {
                return
            }
//            BFLog(1, message: "判断是否开始录音播放** hadPrepareToPlayRecord:\(hadPrepareToPlayRecord), currentPlayRecordIndex:\(currentPlayRecordIndex), isNormalPlaying :\(sself.isNormalPlaying),\(recordPlayer.currentItem?.duration.timescale ?? 0),\(CMTimeGetSeconds(currentT) >= recordedAudio.startCMTime.seconds),\(CMTimeGetSeconds(currentT) <= recordedAudio.endCMTime.seconds - 0.2)")

            if !hadPrepareToPlayRecord,
               CMTimeGetSeconds(currentT) >= (recordedAudio.startCMTime.seconds - 0.1),
               CMTimeGetSeconds(currentT) <= recordedAudio.endCMTime.seconds - 0.2 // 这个条件是避免录音结束后有小幅度回退导致播放最新录音
            {
                if itemModels[currItemModelIndex].mediaType == .Video, recordPlayer.currentItem?.duration.timescale == 0 {
                    return
                }
                // 应当开始播放了
                // 两个逻辑：如果在播，则跳过;如果暂停拖动到中间，则seek
                if currentPlayRecordIndex == -1, sself.isNormalPlaying {
                    let second = currentT - recordedAudio.startCMTime
                    DispatchQueue.main.async { [weak self] in
                        self?.recordPlayer.seek(to: second, toleranceBefore: CMTime(value: 1, timescale: 1_000_000), toleranceAfter: CMTime(value: 1, timescale: 1_000_000), completionHandler: { [weak self] finished in
                            if finished, self?.isNormalPlaying ?? false {
                                self?.recordPlayer.play()
//                                self?.recordPlayer.volume = 1
//                                self?.assetPlayer.volume = self?.haveSpeakVolume ?? 0
                                BFLog(1, message: "录音开始播放 playing, \(second), \(CMTimeGetSeconds(self?.recordPlayer.currentItem?.duration ?? .zero)), \(self?.recordPlayer.currentItem?.currentTime().seconds ?? 0)")
                            }
                        })
                    }
                    currentPlayRecordIndex = shouldPlayRecordIndex
                    hadPrepareToPlayRecord = true
                    BFLog(3, message: "重置播放index-\(#function) = \(currentPlayRecordIndex)")
                    BFLog(1, message: "录音开始播放2, \(second), \(CMTimeGetSeconds(recordPlayer.currentItem?.duration ?? .zero)),index = \(currentPlayRecordIndex)")
                }
            }
        }
    }

    /// 图片素材录音的播放
    /// - Parameters:
    ///   - currentT: <#currentT description#>
    ///   - shouldPlayRecordIndex: <#shouldPlayRecordIndex description#>
    ///   - recordedAudio: <#recordedAudio description#>
    func imageMaterialRecordPlay(at currentT: CMTime, shouldPlayRecordIndex _: Int, recordedAudio: PQVoiceModel) {
        synced(currentPlayRecordIndex) { [weak self] in
            guard let sself = self, sself.isNormalPlaying else {
                return
            }
//            BFLog(1, message: "判断是否开始录音播放** hadPrepareToPlayRecord:\(hadPrepareToPlayRecord), currentPlayRecordIndex:\(currentPlayRecordIndex), isNormalPlaying :\(sself.isNormalPlaying),\(recordPlayer.currentItem?.duration.timescale ?? 0),\(CMTimeGetSeconds(currentT) >= recordedAudio.startCMTime.seconds),\(CMTimeGetSeconds(currentT) <= recordedAudio.endCMTime.seconds - 0.2)")

            if CMTimeGetSeconds(currentT) >= (recordedAudio.startCMTime.seconds - 0.1),
               CMTimeCompare(currentT, recordedAudio.endCMTime) <= 0 // 这个条件是避免录音结束后有小幅度回退导致播放最新录音
            {
                // 应当开始播放了
                // 两个逻辑：如果在播，则跳过;如果暂停拖动到中间，则seek
                if sself.isNormalPlaying {
                    let second = currentT - recordedAudio.startCMTime
                    sself.recordPlayer.seek(to: second, toleranceBefore: CMTime(seconds: 1, preferredTimescale: 1000), toleranceAfter: CMTime(seconds: 1, preferredTimescale: 1000), completionHandler: { isFinished in
                    })
                    sself.recordPlayer.play()
                    BFLog(3, message: "录音开始播放2, \(second), \(CMTimeGetSeconds(recordPlayer.currentItem?.duration ?? .zero)),index = \(currentPlayRecordIndex)")
                }
            }
        }
    }

    func play() {
        BFLog(1, message: "开始播放 \(currentAssetProgress.seconds)")
        recorderManager?.voiceModel = nil

        itemModels[currItemModelIndex].titleStickers.sort { m1, m2 in
            m1.timelineIn < m2.timelineIn
        }

        isNormalPlaying = true
        if isEndPlay || (itemModels[currItemModelIndex].mediaType == .Image && CMTimeCompare(currentAssetProgress, itemModels[currItemModelIndex].materialDuraion) >= 0) {
            isEndPlay = false
            assetPlayer.seek(to: CMTime.zero)
            progressThumV.progress = 0
            currentPlayRecordIndex = -1
            BFLog(3, message: "重置播放index-\(#function) = \(currentPlayRecordIndex)")
            if itemModels[currItemModelIndex].mediaType == .Video {
                recordBtn.isHidden = false
            }
            currentAssetProgress = CMTime.zero
        }
        
        rscurrentManager.play()
        

        deleteRecordBtn.isHidden = true

        subtitleBtn.isHidden = true
        soundSettingBtn.isHidden = true
        withDrawBtn.isHidden = true
        recordBtn.isHidden = true
    }

    func pause() {
        BFLog(1, message: "暂停播放")
        isNormalPlaying = false

        // ---- 修复暂停播放回退问题
        avplayerTimeObserver?.invalidate()
        avplayerTimeObserver = nil
        recordPlayerTimeObserver?.invalidate()
        recordPlayerTimeObserver = nil
        // ----

        subtitleBtn.isHidden = false
        soundSettingBtn.isHidden = false
        withDrawBtn.isHidden = false
        recordBtn.isHidden = (itemModels[currItemModelIndex].mediaType == .Image && isEndPlay) ? false : isEndPlay

        assetPlayer.pause()
        recordPlayer.pause()
        recordStartPlayTime = CMTime.zero
        pauseTime = currentAssetProgress.seconds
        currentPlayRecordIndex = -1
        hadPrepareToPlayRecord = false

        // 暂停状态
        playBtn.isSelected = (itemModels[currItemModelIndex].mediaType == .Image && itemModels[currItemModelIndex].voiceStickers.count <= 0)
        playBtn.isHidden = playBtn.isSelected
    }

    func fetchVideo() {
        if assets.count > 0 {
            for (index, asset) in assets.enumerated() {
                let itemModel = BFRecordItemModel()
                itemModel.index = index
                itemModel.initOriginData(phasset: asset)
                if asset.mediaType != .video {
                    itemModel.coverPath = asset.localPath
                }
                itemModels.append(itemModel)
                if index == 0 {
                    if asset.mediaType == .video {
                        itemModel.fetchAVUrlAssetCallBack = { [weak self, weak itemModel] in
                            DispatchQueue.main.async { [weak self, weak itemModel] in
                                self?.progressThumV.recordItem = itemModel
                                self?.progressThumV.isHidden = false
                                self?.recordBtn.isEnabled = true
                            }
                        }
                    } else {
                        recordBtn.isEnabled = true
                    }
                }
            }
        }else {
            let itemModel = BFRecordItemModel()
            itemModel.index = 0
            itemModel.mediaType = .Camera
            itemModels.append(itemModel)
            
            recordBtn.isEnabled = true
        }
        collectionView.reloadData()

        // 暂停状态--如果是图片素材同时没有录音文件时不显示播放按钮
        playBtn.isSelected = (itemModels.first?.mediaType == .Image && (itemModels.first?.voiceStickers.count ?? 0) <= 0)
        playBtn.isHidden = playBtn.isSelected
    }

    func setCoverImage(img: UIImage) {
        if let pic = GPUImagePicture(image: img) {
            let filter = GPUImageFilter()
            pic.addTarget(filter)
            filter.addTarget(playView)
            pic.processImage()
        }
    }

    func setAudioPlay(item: AVPlayerItem?) {
        BFLog(message: "设置播放器item：\(String(describing: item))")
        guard let item = item else {
            return
        }
        if let playItem = assetPlayer.currentItem {
            NotificationCenter.default.removeObserver(self, name: .AVPlayerItemDidPlayToEndTime, object: playItem)
            assetPlayer.replaceCurrentItem(with: item)
            BFLog(message: "设置播放器playItem-替换：\(String(describing: item))")
        } else {
            assetPlayer.replaceCurrentItem(with: item)
            BFLog(message: "设置播放器item-替换：\(String(describing: item))")
//            assetPlayer.volume = noSpeakVolume
            avplayerTimeObserver?.invalidate()
            avplayerTimeObserver = assetPlayer.addPeriodicTimeObserver(forInterval: CMTime(value: 1, timescale: 1000), queue: DispatchQueue.global()) { [weak self] time in
                // 进度监控
                self?.periodicTimeObserver(item: item, time: time)
                if self?.isNormalPlaying ?? false {
                    // 播放录音
                    self?.startPlayRecord(time: time)
                }
            } as? NSKeyValueObservation
        }

        NotificationCenter.default.addObserver(forName: .AVPlayerItemDidPlayToEndTime, object: assetPlayer.currentItem, queue: .main) { [weak self] _ in
            guard let sself = self else {
                return
            }

            // 非录音的播放状态，结束时自动跳下一段落
            if !sself.isRecording, sself.currItemModelIndex < (sself.itemModels.count - 1) {
//                self?.collectionView.setContentOffset(CGPoint(x: CGFloat((self?.currItemModelIndex ?? 0) + 1) * (self?.collectionView.frame.width ?? 0), y: 0), animated: true)
            } else {} //

            sself.isNormalPlaying = false
            sself.isEndPlay = true
            sself.recordBtn.isHidden = (sself.itemModels[sself.currItemModelIndex].mediaType == .Image) ? false : true

            sself.subtitleBtn.isHidden = false
            sself.soundSettingBtn.isHidden = false

            if sself.isRecording {
                sself.endRecord()
                cShowHUB(superView: nil, msg: "此视频已录制到头了哦")
            }
            sself.playBtn.isSelected = false
            sself.playBtn.isHidden = sself.playBtn.isSelected
        }
    }

    /// 音视频播放进度-视频跟图片单独处理
    /// - Parameter CMTime: <#CMTime description#>
    func periodicTimeObserver(item: AVPlayerItem?, time: CMTime) {
        // 进度监控
        //        if time.seconds < pauseTime {
        //            return
        //        }
        if itemModels[currItemModelIndex].mediaType == .Video {
            if CMTimeGetSeconds(item?.duration ?? CMTime.zero) > 0 {
                currentAssetProgress = CMTime(seconds: time.seconds, preferredTimescale: 1000)
//                BFLog(1, message: "video curr:\(CMTimeGetSeconds(currentAssetProgress))")
                DispatchQueue.main.async { [weak self] in
                    BFLog(message: "更新录音进度\(#function)-\(self?.currentAssetProgress.seconds ?? 0)")
                    self?.progreddL.text = String(format: "%@", CMTimeGetSeconds(time).formatDurationToHMS())
                    let su = !(self?.isDragingProgressSlder ?? false) || (self?.isRecording ?? false) || (self?.isNormalPlaying ?? false)
                    if su { // 不拖动，正常播放和录音时更新进度条
                        self?.progressThumV.progress = time.seconds
                    }
                    // 更新字幕
                    if !(self?.isRecording ?? false) {
                        self?.updateSubtitle(time: time)
                    }
                }
            }
        }

        if isRecording {
            let startTime = recorderManager?.voiceModel?.startCMTime.seconds ?? 0
            let progress = (currentAssetProgress - (recorderManager?.voiceModel?.startCMTime ?? .zero)).seconds // - ratioX
            // 使用播放器的进度来画线，因为进度是跟着播放器来了
            DispatchQueue.main.async { [weak self] in
                self?.indirectionView?.setProgress(start: startTime, progress: max(0, progress))
            }
        }
    }

    // MARK: - 录音对应图像绘制

    // 撤销按钮修改title，重绘
    func changeWithDrawBtnLayout(_ type: Int) {
        switch type {
        case 3:
            withDrawBtn.setTitle("撤销删除", for: .normal)
        case 2:
            withDrawBtn.setTitle("撤销录制", for: .normal)
        default:
            withDrawBtn.setTitle("回退", for: .normal)
        }
        withDrawBtn.imageEdgeInsets = UIEdgeInsets(top: -withDrawBtn.imageView!.height, left: 0, bottom: 0, right: -withDrawBtn.titleLabel!.width)
        withDrawBtn.titleEdgeInsets = UIEdgeInsets(top: withDrawBtn.titleLabel!.height + 2, left: -withDrawBtn.imageView!.width, bottom: 0, right: 0)

        withDrawBtn.setNeedsLayout()
        withDrawBtn.layoutIfNeeded()
    }

    // 通过缩略图进度条控制播放进度
    // progress : 图片且isback为true时，表示是时长；  其他情况： 0 - 1，百分比
    func changeProgress(isBack: Bool = false, progress: Float = -1, changCMTime: CMTime = .zero) {
        var newProgress = progress
        if progress.isNaN || progress.isInfinite {
            newProgress = 0
        }
        if itemModels[currItemModelIndex].mediaType == .Video {
            let duration = itemModels[currItemModelIndex].materialDuraion
            if duration.seconds > 0 {
                if progress == -1 {
                    currentAssetProgress = changCMTime
                } else {
                    currentAssetProgress = CMTime(seconds: Double(newProgress) * duration.seconds, preferredTimescale: 1000)
                }
                DispatchQueue.main.async { [weak self] in
                    BFLog(message: "更新录音进度\(#function)- \(self?.currentAssetProgress.seconds ?? 0)")
                    self!.progreddL.text = String(format: "%@", CMTimeGetSeconds(self!.currentAssetProgress).formatDurationToHMS())
                }
                assetPlayer.seek(to: currentAssetProgress, toleranceBefore: CMTime(value: 1, timescale: 1_000_000), toleranceAfter: CMTime(value: 1, timescale: 1_000_000)) { _ in
                }
            }
        } else {
            currentAssetProgress = isBack ? CMTime(value: CMTimeValue(newProgress * 1000), timescale: 1000) : (progress != -1 ? CMTime(value: CMTimeValue(newProgress * Float(itemModels[currItemModelIndex].materialDuraion.seconds) * 1000), timescale: 1000) : changCMTime)
            if currentAssetProgress.seconds > itemModels[currItemModelIndex].materialDuraion.seconds {
                currentAssetProgress = itemModels[currItemModelIndex].materialDuraion
            }
            DispatchQueue.main.async { [weak self] in
                BFLog(message: "更新录音进度\(#function)- \(self?.currentAssetProgress.seconds ?? 0)")
                self?.progreddL.text = String(format: "%@", CMTimeGetSeconds(self!.currentAssetProgress).formatDurationToHMS())
            }
            BFLog(message: "progress = \(progress),currentAssetProgress = \(currentAssetProgress.seconds),materialDuraion = \(itemModels[currItemModelIndex].materialDuraion.seconds)")

            // add by ak 图片素材拖动指针时同时更新字幕显现数据
            updateSubtitle(time: currentAssetProgress)
        }
    }

    // 画进度线
    func drawProgressIndication(progress: Double) {
        if indirectionView == nil {
            var percenWidth: CGFloat = 0
            if itemModels[currItemModelIndex].mediaType == .Image {
                percenWidth = progressThumV.thumbImageWidth / 2.0
            } else {
                percenWidth = progressThumV.progessIndicateBackV.frame.width / CGFloat(itemModels[currItemModelIndex].materialDuraion.seconds)
            }
            indirectionView = BFIndirectionProgressView(frame: CGRect(origin: progressThumV.progessIndicateBackV.origin, size: CGSize(width: progressThumV.progressView.contentSize.width, height: progressThumV.progessIndicateBackV.frame.height)), percenWidth: percenWidth, totalDuration: itemModels[currItemModelIndex].materialDuraion.seconds)
            progressThumV.progressView.addSubview((indirectionView)!)
        }
        // 更新录制进度
        // 注：视频无法以录制进度驱动，因当录音开始录音时播放器还未播放，导致进度不一致
        // 注：在录制停止时，视频播放器进度依然在走，误差在80毫秒左右
        if isRecording, itemModels[currItemModelIndex].mediaType == .Image {
            let startTime = recorderManager?.voiceModel?.startCMTime.seconds ?? 0
            // 使用播放器的进度来画线，因为进度是跟着播放器来了
            indirectionView?.setProgress(start: startTime, progress: progress)
        }
        if itemModels[currItemModelIndex].mediaType == .Image {
            imageRecordProgress(isRecord: true, progress: progress)
        }
    }
}

extension BFRecordScreenController: GPUImageMovieDelegate {
    public func didCompletePlayingMovie() {
        BFLog(1, message: "播放结束")
        currentPlayRecordIndex = -1
        BFLog(3, message: "重置播放index-\(#function) = \(currentPlayRecordIndex)")
    }
}

extension BFRecordScreenController: AVAudioRecorderDelegate {
    public func audioRecorderDidFinishRecording(_: AVAudioRecorder, successfully _: Bool) {
        BFLog(1, message: "录音结束")
    }
}

extension BFRecordScreenController: AVAudioPlayerDelegate {
    public func audioPlayerDidFinishPlaying(_: AVAudioPlayer, successfully _: Bool) {
        BFLog(1, message: "录音播放结束")
    }
}

// MARK: - UICollectionViewDelegate

/// UICollectionViewDelegate
extension BFRecordScreenController: UICollectionViewDelegate, UICollectionViewDataSource, UIScrollViewDelegate {
    public func collectionView(_: UICollectionView, numberOfItemsInSection _: Int) -> Int {
        return itemModels.count
    }

    public func collectionView(_ collectionView: UICollectionView, cellForItemAt indexPath: IndexPath) -> UICollectionViewCell {
        let recordItem = itemModels[indexPath.item]
        var cell: BFImageCoverViewCell!

        switch recordItem.mediaType {
        case .Image:
            cell = BFImageCoverViewCell.gpuImageViewCell(collectionView: collectionView, indexPath: indexPath)
        case .Video:
            cell = BFVideoCoverViewCell.gpuVideoViewCell(collectionView: collectionView, indexPath: indexPath)
        case .Camera:
            cell = BFCameraCoverViewCell.gpuCamraViewCell(collectionView: collectionView, indexPath: indexPath)
            (cell as! BFCameraCoverViewCell).initCameraAfterData = {[weak self] in
//                if self?.currItemModelIndex == indexPath.row {
//                    // 开启摄像头
//                    self?.rscurrentManager.playView = cell.playView
//                    self?.rscurrentManager.resetEnv()
//                }
            }
        default:
            break
        }
        
        recordItem.fetchCoverImgCallBack = { [weak self, weak cell, weak recordItem] _ in
            guard let sself = self, let item = recordItem else {
                return
            }
            cell?.addData()
            if item.mediaType == .Image {
                sself.progressThumV.recordItem = item
                sself.progressThumV.isHidden = false
            }
        }
        recordItem.fetchPlayItemCallBack = { [weak self, weak recordItem] item in
            guard let sself = self else {
                return
            }
            guard item != nil else {
                cShowHUB(superView: nil, msg: "视频获取失败:\(recordItem?.index ?? 0)")
                return
            }
            sself.rscurrentManager.resetEnv()
        }
        cell.btnClickHandle = { [weak self] _, _ in
            guard let sself = self else {
                return
            }
            sself.playVideo(btn: sself.playBtn)
        }
        cell.recordItem = recordItem
        return cell
    }

    public func collectionView(_: UICollectionView, didSelectItemAt _: IndexPath) {}

    public func scrollViewWillBeginDragging(_: UIScrollView) {
        BFLog(1, message: "开始滚动")
        recordBtn.isEnabled = false
    }

    public func scrollViewDidEndDecelerating(_ scrollView: UIScrollView) {
        endScrollItem(scrollView)
    }

    public func scrollViewDidEndDragging(_ scrollView: UIScrollView, willDecelerate decelerate: Bool) {
        if !decelerate {
            endScrollItem(scrollView)
        }
    }

    func endScrollItem(_ scrollView: UIScrollView) {
//    public func scrollViewDidScroll(_ scrollView: UIScrollView) {
        BFLog(1, message: "滚动结束")

        let page = Int((scrollView.contentOffset.x + scrollView.frame.width / 2) / scrollView.frame.width)
        if page != currItemModelIndex {
            // 切换素材时先把录制状态切为不可用,延迟可点，避免在缩略图未加载出来时即可录制
            currItemModelIndex = page
            // 暂停
            pause()
            // 如果在录制中，停止录制
            if isRecording {
                endRecord()
            }
            recorderManager?.cancelTitleService()
//            events.append(WithDrawModel(type: 0, timestamp: currentAssetProgress))
//            // 重设撤销栈
//            itemModels[currItemModelIndex].events = events
//            events = itemModels[page].events
            events = [WithDrawModel]()


            let recordItem = itemModels[currItemModelIndex]
            // 暂停状态--如果是图片素材同时没有录音文件时不显示播放按钮
            playBtn.isSelected = (recordItem.mediaType == .Image && recordItem.voiceStickers.count <= 0)
            playBtn.isHidden = playBtn.isSelected
            // 重绘录音区域
            progressThumV.recordItem = recordItem
            DispatchQueue.main.asyncAfter(deadline: .now() + 0.1) { [weak self] in
                /// 重绘录音进度视图
                self?.resetAllIndirectionView()
                // 重绘录音进度
                self?.recordBtn.isEnabled = true
            }
            // 更新缩略图
            progressThumV.isHidden = false
            progreddL.text = "00:00"
            // 重置指针
            currentAssetProgress = CMTime(seconds: 0, preferredTimescale: 1000)
            searchStopAtRecordRange()
            // 切换要更新当前录制index，避免在录制完以后切换素材这种时候为-3会拦截
            currentPlayRecordIndex = -1
            showSubtitleIndex = -1
            // 重置播放器
            assetPlayer.seek(to: CMTime.zero)
            recordPlayer.seek(to: CMTime.zero)

            if let voice = itemModels[page].voiceStickers.enumerated().first(where: { m in
                m.1.startTime == 0
            }) {
                currentPlayRecordIndex = voice.0
            }

            searchStopAtRecordRange()
            changeWithDrawBtnLayout(0)
            pauseTime = 0

            if changeItemHandle != nil {
                changeItemHandle!(page)
            }

            // add by ak 切换后刷新字幕
            subtitleLabel.removeFromSuperview()
            let cell = collectionView.cellForItem(at: IndexPath(item: page, section: 0)) as? BFImageCoverViewCell
            cell?.contentView.addSubview(subtitleLabel)
            updateSubtitle(time: currentAssetProgress)
        }
        recordBtn.isEnabled = true
    }

    public func updateContentOffset(index: Int) {
        let offx = collectionView.contentOffset.x
        if offx != collectionView.frame.width * CGFloat(index) {
            collectionView.setContentOffset(CGPoint(x: collectionView.frame.width * CGFloat(index), y: 0), animated: true)
        }
    }
}

// MARK: - 处理图片素材

/// 处理图片素材
public extension BFRecordScreenController {
    /// 播放录音
    /// - Parameter time: <#time description#>
    func startPlayRecord(time: CMTime) {
        // 播放对应的录音音频
        if itemModels[currItemModelIndex].mediaType == .Image {
            if itemModels[currItemModelIndex].materialDuraion.seconds <= 0 {
                playBtn.isSelected = true
                playBtn.isHidden = playBtn.isSelected
                return
            }
            isNormalPlaying = true
            // 当开始播放时重置录音播放起始时间
            recordStartPlayTime = currentAssetProgress
            currenStartPlayTime = CMTime.zero
        }
        playRecord(at: time, periodicTimeObserver: { [weak self] currentT, _ in
//            BFLog(1, message: "播放录音进度：\(currentT.seconds),\(currentItem)")
            if self?.itemModels[self?.currItemModelIndex ?? 0].mediaType == .Image, self?.isNormalPlaying ?? false {
                self?.imageRecordProgress(progress: CMTimeGetSeconds(currentT))
            }
        }, didPlayToEndTime: { [weak self] recordInfo, currentItem in
            BFLog(message: "播放录音结束：\(String(describing: recordInfo?.1)),\(String(describing: currentItem))")
            if self?.itemModels[self?.currItemModelIndex ?? 0].mediaType == .Image {
                if (self?.itemModels[self?.currItemModelIndex ?? 0].voiceStickers.count ?? 0) <= ((recordInfo?.0 ?? 0) + 1) || CMTimeCompare((recordInfo?.1.endCMTime ?? .zero), (self?.itemModels[self?.currItemModelIndex ?? 0].voiceStickers.last?.endCMTime ?? .zero)) >= 0 {
                    
                    self?.isEndPlay = true
                    self?.pause()
                    // 注：矫正进度--播放结束后当前指针应该到当前素材总时长
                    self?.currentAssetProgress = CMTime(seconds: self?.itemModels[self?.currItemModelIndex ?? 0].materialDuraion.seconds ?? 0, preferredTimescale: 1000)
                    self?.resetCurrentProgress()
                    // 录制播放结束后显示录制按钮
                    self?.recordBtn.isHidden = false
                    self?.recordBtn.alpha = 1
                } else {
                    // 注：矫正进度--一段录音播放结束后当前指针应该到当前录音结束点
                    self?.currentAssetProgress = self?.itemModels[self?.currItemModelIndex ?? 0].voiceStickers[(recordInfo?.0 ?? 0) + 1].startCMTime ?? .zero
                    self?.startPlayRecord(time: self?.currentAssetProgress ?? CMTime.zero)
                }
            }
        }) { [weak self] _, _ in
            if self?.itemModels[self?.currItemModelIndex ?? 0].mediaType == .Image {
                DispatchQueue.main.asyncAfter(deadline: DispatchTime.now() + 3) {
                    self?.startPlayRecord(time: self?.currentAssetProgress ?? CMTime.zero)
                }
            }
        }
    }

    /// 处理图片素材录音
    func imageRecordProgress(isRecord: Bool = false, progress: Float64) {
        if isRecord {
            currentAssetProgress = (recorderManager?.voiceModel?.startCMTime ?? .zero) + CMTime(seconds: progress, preferredTimescale: 1000)
        } else {
            currentAssetProgress = recordStartPlayTime + CMTime(seconds: progress, preferredTimescale: 1000)
        }
        BFLog(1, message: "图片录音进度：\(progress),currentAssetProgress=\(currentAssetProgress.seconds),\(itemModels[currItemModelIndex].materialDuraion)")
        if itemModels[currItemModelIndex].mediaType == .Image {
            /// 重置进度
            resetCurrentProgress()
        }
    }

    /// 重置进度
    func resetCurrentProgress() {
        DispatchQueue.main.async { [weak self] in
            if !(self?.isRecording ?? false), CMTimeCompare((self?.currentAssetProgress ?? .zero), (self?.itemModels[self?.currItemModelIndex ?? 0].materialDuraion ?? .zero)) > 0 {
                return
            }
//            BFLog(1, message: "更新录音进度\(#function)-\(self?.currentAssetProgress.seconds ?? 0)")
            self?.progreddL.text = String(format: "%@", (self?.currentAssetProgress.seconds ?? 0).formatDurationToHMS())
            self?.progressThumV.progress = (self?.currentAssetProgress.seconds ?? 0)
            self?.updateSubtitle(time: self?.currentAssetProgress ?? CMTime.zero)
        }
    }

    /// 重绘录音进度视图
    func resetAllIndirectionView() {
        // 重绘录音进度视图
        var percenWidth: CGFloat = 0
        if itemModels[currItemModelIndex].mediaType == .Image {
            percenWidth = progressThumV.thumbImageWidth / 2.0
        } else {
            percenWidth = progressThumV.progessIndicateBackV.frame.width / CGFloat(itemModels[currItemModelIndex].materialDuraion.seconds)
        }
        indirectionView?.resetAllSubViews(items: itemModels[currItemModelIndex].voiceStickers, percenWidth: percenWidth, totalDuration: itemModels[currItemModelIndex].materialDuraion.seconds)
    }
}
